import logging
from typing import TYPE_CHECKING, Optional

import aiosignald.exc as sigexc
import aiosignald.generated as sigapi
from slixmpp.exceptions import XMPPError

from slidge import *

if TYPE_CHECKING:
    from .session import Session


class Contact(LegacyContact["Session"]):
    CORRECTION = False

    def __init__(
        self,
        session: "Session",
        phone: str,
        jid_username: str,
    ):
        super().__init__(session, phone, jid_username)
        log.debug("JID: %s", self.jid_username)
        self._uuid: Optional[str] = None

    @property
    def phone(self):
        return self.legacy_id

    @phone.setter
    def phone(self, p):
        if p is not None:
            self.session.contacts.contacts_by_legacy_id[p] = self
            self.legacy_id = p
            self.jid_username = p

    @property
    def uuid(self):
        return self._uuid

    @uuid.setter
    def uuid(self, u: str):
        if u is not None:
            log.debug("UUID: %s, %s", u, self)
            self.session.contacts.contacts_by_uuid[u] = self
        self._uuid = u

    @property
    def signal_address(self):
        return sigapi.JsonAddressv1(number=self.phone, uuid=self.uuid)

    async def get_identities(self):
        s = await self.session.signal
        log.debug("%s, %s", type(self.session.phone), type(self.signal_address))
        try:
            r = await s.get_identities(
                account=self.session.phone,
                address=self.signal_address,
            )
        except sigexc.UnregisteredUserError:
            raise XMPPError("not-found")
        identities = r.identities
        self.session.send_gateway_message(str(identities))


class Roster(LegacyRoster[Contact, "Session"]):
    def __init__(self, session):
        super().__init__(session)
        self.contacts_by_uuid: dict[str, Contact] = {}
        self.contacts_by_legacy_id = self._contacts_by_legacy_id
        self.contacts_by_bare_jid = self._contacts_by_bare_jid

    def by_jid(self, contact_jid):
        if (c := self.contacts_by_legacy_id.get(contact_jid.user)) is None:
            return super().by_jid(contact_jid)
        else:
            return c

    def by_phone(self, phone: str):
        return self.by_legacy_id(phone)

    async def by_uuid(self, uuid: str):
        try:
            return self.contacts_by_uuid[uuid]
        except KeyError:
            address = await (await self.session.signal).resolve_address(
                account=self.session.phone, partial=sigapi.JsonAddressv1(uuid=uuid)
            )
            return await self.by_json_address(address)

    async def by_json_address(self, address: sigapi.JsonAddressv1):
        uuid = address.uuid
        phone = address.number

        if uuid is None and phone is None:
            raise TypeError(address)

        if uuid is None:
            return self.by_phone(phone)

        if phone is None:
            return await self.by_uuid(uuid)

        contact_phone = self._contacts_by_legacy_id.get(phone)
        contact_uuid = self.contacts_by_uuid.get(uuid)

        if contact_phone is None and contact_uuid is None:
            c = self.by_phone(phone)
            c.uuid = uuid
            return c

        if contact_phone is None and contact_uuid is not None:
            contact_uuid.phone = phone
            return contact_uuid

        if contact_uuid is None and contact_phone is not None:
            contact_phone.uuid = uuid
            return contact_phone

        if contact_phone is not contact_uuid:
            raise RuntimeError(address, contact_phone, contact_uuid)

        return contact_phone


log = logging.getLogger(__name__)




# Slixmpp: The Slick XMPP Library
# Copyright (C) 2010 Nathanael C. Fritz, Lance J.T. Stout
# This file is part of Slixmpp.
# See the file LICENSE for copying permission.
from slixmpp.plugins.base import register_plugin

from . import stanza
from .disco import XEP_0030
from .stanza import DiscoInfo, DiscoItems
from .static import StaticDisco

register_plugin(XEP_0030)



# Slixmpp: The Slick XMPP Library
# Copyright (C) 2011 Nathanael C. Fritz, Lance J.T. Stout
# This file is part of Slixmpp.
# See the file LICENSE for copying permission.
import base64
import hashlib
import logging
from asyncio import Future
from typing import Optional

from slixmpp import __version__
from slixmpp.exceptions import XMPPError
from slixmpp.plugins import BasePlugin
from slixmpp.stanza import Iq, Presence, StreamFeatures
from slixmpp.types import OptJidStr
from slixmpp.util import MemoryCache
from slixmpp.xmlstream import JID, register_stanza_plugin
from slixmpp.xmlstream.handler import Callback
from slixmpp.xmlstream.matcher import StanzaPath

from . import StaticCaps, stanza

log = logging.getLogger(__name__)


class XEP_0115(BasePlugin):

    """
    XEP-0115: Entity Capabilities
    """

    name = "xep_0115"
    description = "XEP-0115: Entity Capabilities"
    dependencies = {"xep_0030", "xep_0128", "xep_0004"}
    stanza = stanza
    default_config = {
        "hash": "sha-1",
        "caps_node": None,
        "broadcast": True,
        "cache": None,
    }

    def plugin_init(self):
        self.hashes = {"sha-1": hashlib.sha1, "sha1": hashlib.sha1, "md5": hashlib.md5}

        if self.caps_node is None:
            self.caps_node = "http://slixmpp.com/ver/%s" % __version__

        if self.cache is None:
            self.cache = MemoryCache()

        register_stanza_plugin(Presence, stanza.Capabilities)
        register_stanza_plugin(StreamFeatures, stanza.Capabilities)

        self._disco_ops = [
            "cache_caps",
            "get_caps",
            "assign_verstring",
            "get_verstring",
            "supports",
            "has_identity",
        ]

        self.xmpp.register_handler(
            Callback(
                "Entity Capabilites", StanzaPath("presence/caps"), self._handle_caps
            )
        )

        self.xmpp.add_filter("out", self._filter_add_caps)

        self.xmpp.add_event_handler("entity_caps", self._process_caps)

        if not self.xmpp.is_component:
            self.xmpp.register_feature(
                "caps", self._handle_caps_feature, restart=False, order=10010
            )

        disco = self.xmpp["xep_0030"]
        self.static = StaticCaps(self.xmpp, disco.static)

        for op in self._disco_ops:
            self.api.register(getattr(self.static, op), op, default=True)

        for op in ("supports", "has_identity"):
            self.xmpp["xep_0030"].api.register(getattr(self.static, op), op)

        self._run_node_handler = disco._run_node_handler

        disco.cache_caps = self.cache_caps
        disco.update_caps = self.update_caps
        disco.assign_verstring = self.assign_verstring
        disco.get_verstring = self.get_verstring

    def plugin_end(self):
        self.xmpp["xep_0030"].del_feature(feature=stanza.Capabilities.namespace)
        self.xmpp.del_filter("out", self._filter_add_caps)
        self.xmpp.del_event_handler("entity_caps", self._process_caps)
        self.xmpp.remove_handler("Entity Capabilities")
        if not self.xmpp.is_component:
            self.xmpp.unregister_feature("caps", 10010)
        for op in ("supports", "has_identity"):
            self.xmpp["xep_0030"].restore_defaults(op)

    def session_bind(self, jid):
        self.xmpp["xep_0030"].add_feature(stanza.Capabilities.namespace)

    async def _filter_add_caps(self, stanza):
        if not isinstance(stanza, Presence) or not self.broadcast:
            return stanza

        if stanza["type"] not in ("available", "chat", "away", "dnd", "xa"):
            return stanza

        ver = await self.get_verstring(stanza["from"])
        if ver:
            stanza["caps"]["node"] = self.caps_node
            stanza["caps"]["hash"] = self.hash
            stanza["caps"]["ver"] = ver
        return stanza

    def _handle_caps(self, presence):
        if not self.xmpp.is_component:
            if presence["from"] == self.xmpp.boundjid:
                return
        self.xmpp.event("entity_caps", presence)

    def _handle_caps_feature(self, features):
        # We already have a method to process presence with
        # caps, so wrap things up and use that.
        p = Presence()
        p["from"] = self.xmpp.boundjid.domain
        p.append(features["caps"])
        self.xmpp.features.add("caps")

        self.xmpp.event("entity_caps", p)

    async def _process_caps(self, pres: Presence):
        if not pres["caps"]["hash"]:
            log.debug(
                "Received unsupported legacy caps: %s, %s, %s",
                pres["caps"]["node"],
                pres["caps"]["ver"],
                pres["caps"]["ext"],
            )
            self.xmpp.event("entity_caps_legacy", pres)
            return

        if self.xmpp.is_component and pres.get_to() != self.xmpp.boundjid.bare:
            log.debug("Not attempting to update caps for presence not directed at the component JID")
            return

        ver = pres["caps"]["ver"]

        existing_verstring = await self.get_verstring(pres["from"].full)
        if str(existing_verstring) == str(ver):
            return

        existing_caps = await self.get_caps(verstring=ver)
        if existing_caps is not None:
            await self.assign_verstring(pres["from"], ver)
            return

        ifrom = pres["to"] if self.xmpp.is_component else None

        if pres["caps"]["hash"] not in self.hashes:
            try:
                log.debug("Unknown caps hash: %s", pres["caps"]["hash"])
                self.xmpp["xep_0030"].get_info(jid=pres["from"], ifrom=ifrom)
                return
            except XMPPError:
                return

        log.debug("New caps verification string: %s", ver)
        try:
            node = "%s#%s" % (pres["caps"]["node"], ver)
            caps = await self.xmpp["xep_0030"].get_info(pres["from"], node, ifrom=ifrom)

            if isinstance(caps, Iq):
                caps = caps["disco_info"]

            if await self._validate_caps(
                caps, pres["caps"]["hash"], pres["caps"]["ver"]
            ):
                await self.assign_verstring(pres["from"], pres["caps"]["ver"])
        except XMPPError:
            log.debug("Could not retrieve disco#info results for caps for %s", node)

    async def _validate_caps(self, caps, hash, check_verstring):
        # Check Identities
        full_ids = caps.get_identities(dedupe=False)
        deduped_ids = caps.get_identities()
        if len(full_ids) != len(deduped_ids):
            log.debug("Duplicate disco identities found, invalid for caps")
            return False

        # Check Features
        full_features = caps.get_features(dedupe=False)
        deduped_features = caps.get_features()
        if len(full_features) != len(deduped_features):
            log.debug("Duplicate disco features found, invalid for caps")
            return False

        # Check Forms
        form_types = []
        deduped_form_types = set()
        for stanza in caps["substanzas"]:
            if not isinstance(stanza, self.xmpp["xep_0004"].stanza.Form):
                log.debug("Non form extension found, ignoring for caps")
                caps.xml.remove(stanza.xml)
                continue
            if "FORM_TYPE" in stanza.get_fields():
                f_type = tuple(stanza.get_fields()["FORM_TYPE"]["value"])
                form_types.append(f_type)
                deduped_form_types.add(f_type)
                if len(form_types) != len(deduped_form_types):
                    log.debug("Duplicated FORM_TYPE values, " + "invalid for caps")
                    return False

                if len(f_type) > 1:
                    deduped_type = set(f_type)
                    if len(f_type) != len(deduped_type):
                        log.debug("Extra FORM_TYPE data, invalid for caps")
                        return False

                if stanza.get_fields()["FORM_TYPE"]["type"] != "hidden":
                    log.debug(
                        "Field FORM_TYPE type not 'hidden', " + "ignoring form for caps"
                    )
                    caps.xml.remove(stanza.xml)
            else:
                log.debug("No FORM_TYPE found, ignoring form for caps")
                caps.xml.remove(stanza.xml)

        verstring = self.generate_verstring(caps, hash)
        if verstring != check_verstring:
            log.debug(
                "Verification strings do not match: %s, %s"
                % (verstring, check_verstring)
            )
            return False

        await self.cache_caps(verstring, caps)
        return True

    def generate_verstring(self, info, hash):
        hash = self.hashes.get(hash, None)
        if hash is None:
            return None

        S = ""

        # Convert None to '' in the identities
        def clean_identity(id):
            return map(lambda i: i or "", id)

        identities = map(clean_identity, info["identities"])

        identities = sorted(("/".join(i) for i in identities))
        features = sorted(info["features"])

        S += "<".join(identities) + "<"
        S += "<".join(features) + "<"

        form_types = {}

        for stanza in info["substanzas"]:
            if isinstance(stanza, self.xmpp["xep_0004"].stanza.Form):
                if "FORM_TYPE" in stanza.get_fields():
                    f_type = stanza["values"]["FORM_TYPE"]
                    if len(f_type):
                        f_type = f_type[0]
                    if f_type not in form_types:
                        form_types[f_type] = []
                    form_types[f_type].append(stanza)

        sorted_forms = sorted(form_types.keys())
        for f_type in sorted_forms:
            for form in form_types[f_type]:
                S += "%s<" % f_type
                fields = sorted(form.get_fields().keys())
                fields.remove("FORM_TYPE")
                for field in fields:
                    S += "%s<" % field
                    vals = form.get_fields()[field].get_value(convert=False)
                    if vals is None:
                        S += "<"
                    else:
                        if not isinstance(vals, list):
                            vals = [vals]
                        S += "<".join(sorted(vals)) + "<"

        binary = hash(S.encode("utf8")).digest()
        return base64.b64encode(binary).decode("utf-8")

    async def update_caps(
        self,
        jid: OptJidStr = None,
        node: Optional[str] = None,
        preserve: bool = False,
        broadcast: bool = True,
    ):
        """Update caps for a local JID based on current data.

        :param jid: JID whose info to update
        :param node: Node to fetch info from
        :param broadcast: Send a presence after updating.
        :param preserve: Send presence only to contacts found in the roster.
        """
        try:
            info = await self.xmpp["xep_0030"].get_info(jid, node, local=True)
            if isinstance(info, Iq):
                info = info["disco_info"]
            ver = self.generate_verstring(info, self.hash)
            await self.xmpp["xep_0030"].set_info(
                jid=jid, node="%s#%s" % (self.caps_node, ver), info=info
            )
            await self.cache_caps(ver, info)
            await self.assign_verstring(jid, ver)

            if broadcast and self.xmpp.sessionstarted and self.broadcast:
                if self.xmpp.is_component or preserve:
                    for contact in self.xmpp.roster[jid]:
                        self.xmpp.roster[jid][contact].send_last_presence()
                else:
                    self.xmpp.roster[jid].send_last_presence()
        except XMPPError:
            return

    def get_verstring(self, jid=None) -> Future:
        """Get the stored verstring for a JID.

        .. versionchanged:: 1.8.0
            This function now returns a Future.
        """
        if jid in ("", None):
            jid = self.xmpp.boundjid.full
        if isinstance(jid, JID):
            jid = jid.full
        return self.api["get_verstring"](jid)

    def assign_verstring(self, jid=None, verstring=None) -> Future:
        """Assign a vertification string to a jid.

        .. versionchanged:: 1.8.0
            This function now returns a Future.
        """
        if jid in (None, ""):
            jid = self.xmpp.boundjid.full
        if isinstance(jid, JID):
            jid = jid.full
        return self.api["assign_verstring"](jid, args={"verstring": verstring})

    def cache_caps(self, verstring=None, info=None) -> Future:
        """Add caps to the cache.

        .. versionchanged:: 1.8.0
            This function now returns a Future.
        """
        data = {"verstring": verstring, "info": info}
        return self.api["cache_caps"](args=data)

    async def get_caps(self, jid=None, verstring=None):
        """Get caps for a JID.

        .. versionchanged:: 1.8.0
            This function is now a coroutine.
        """
        if verstring is None:
            if jid is not None:
                verstring = await self.get_verstring(jid)
            else:
                return None
        if isinstance(jid, JID):
            jid = jid.full
        data = {"verstring": verstring}
        return await self.api["get_caps"](jid, args=data)



import asyncio
import datetime
import logging
from typing import TYPE_CHECKING

import aiotdlib.api as tgapi

from slidge import *

if TYPE_CHECKING:
    from .session import Session


class Contact(LegacyContact["Session"]):
    legacy_id: int
    # Telegram official clients have no XMPP presence equivalent, but a 'last seen' indication.
    AWAY_DELAY = 300

    def __init__(self, *a, **k):
        super(Contact, self).__init__(*a, **k)
        self.last_seen = datetime.datetime.fromtimestamp(0)
        self.away_task: asyncio.Task = self.xmpp.loop.create_task(self.delayed_away())

    def reset_delayed_away(self):
        self.away_task.cancel()
        self.last_seen = datetime.datetime.now()
        self.online()
        self.away_task = self.xmpp.loop.create_task(self.delayed_away())

    async def delayed_away(self):
        await asyncio.sleep(60)
        for x in range(1, 60):
            self.away(f"Last seen {x} minute{'s' if x > 1 else ''} ago")
            await asyncio.sleep(60)
        for x in range(1, 24):
            self.away(f"Last seen {x} hour{'s' if x > 1 else ''} ago")
            await asyncio.sleep(3600)
        self.away(f"Last seen yesterday")
        await asyncio.sleep(3600 * 24)
        for x in range(2, 7):
            self.away(f"Last seen {x} day{'s' if x > 1 else ''} ago")
            await asyncio.sleep(3600 * 24)
        self.away()

    def away(self, status=None):
        super().away(status or f"Last seen: {self.last_seen:%%B %d, %Y}")

    def active(self):
        self.reset_delayed_away()
        self.online()
        super().active()

    async def send_tg_message(self, msg: tgapi.Message):
        self.reset_delayed_away()
        content = msg.content
        if isinstance(content, tgapi.MessageText):
            # TODO: parse formatted text to markdown
            formatted_text = content.text
            self.send_text(
                body=formatted_text.text,
                legacy_msg_id=msg.id,
                reply_to_msg_id=msg.reply_to_message_id,
            )
        elif isinstance(content, tgapi.MessageAnimatedEmoji):
            emoji = content.animated_emoji.sticker.emoji
            self.send_text(
                body=emoji,
                legacy_msg_id=msg.id,
                reply_to_msg_id=msg.reply_to_message_id,
            )
        elif isinstance(content, tgapi.MessagePhoto):
            photo = content.photo
            best_file = max(photo.sizes, key=lambda x: x.width).photo
            await self.send_tg_file(best_file, content.caption, msg.id)
        elif isinstance(content, tgapi.MessageVideo):
            best_file = content.video.video
            await self.send_tg_file(best_file, content.caption, msg.id)
        elif isinstance(content, tgapi.MessageAnimation):
            best_file = content.animation.animation
            await self.send_tg_file(best_file, content.caption, msg.id)
        else:
            self.session.log.debug("Ignoring content: %s", type(content))

    async def send_tg_file(self, best_file, caption, msg_id):
        self.reset_delayed_away()
        query = tgapi.DownloadFile.construct(
            file_id=best_file.id, synchronous=True, priority=1
        )
        best_file_downloaded: tgapi.File = await self.session.tg.request(query)
        await self.send_file(best_file_downloaded.local.path)
        if caption.text:
            self.send_text(caption.text, legacy_msg_id=msg_id)

    async def send_tg_status(self, status: tgapi.UserStatus):
        self.reset_delayed_away()
        if isinstance(status, tgapi.UserStatusOnline):
            self.active()
        elif isinstance(status, tgapi.UserStatusOffline):
            self.paused()
            self.inactive()
        else:
            log.debug("Ignoring status %s", status)


class Roster(LegacyRoster["Contact", "Session"]):
    @staticmethod
    def jid_username_to_legacy_id(jid_username: str) -> int:
        return int(jid_username)


log = logging.getLogger(__name__)



import functools
import logging
from datetime import datetime, timezone
from io import BytesIO
from pathlib import Path
from typing import (
    IO,
    TYPE_CHECKING,
    Any,
    Generic,
    Iterable,
    Literal,
    Optional,
    Type,
    TypeVar,
    Union,
)

import aiohttp
from slixmpp import JID, Iq, Message

from ..util import SubclassableOnce
from ..util.types import (
    AvatarType,
    LegacyContactIdType,
    LegacyMessageType,
    LegacyUserIdType,
)

if TYPE_CHECKING:
    from .session import SessionType
else:
    SessionType = TypeVar("SessionType")


class LegacyContact(Generic[SessionType], metaclass=SubclassableOnce):
    """
    This class centralizes actions in relation to a specific legacy contact.

    You shouldn't create instances of contacts manually, but rather rely on
    :meth:`.LegacyRoster.by_legacy_id` to ensure that contact instances are
    singletons. The :class:`.LegacyRoster` instance of a session is accessible
    through the :attr:`.BaseSession.contacts` attribute.

    Typically, your plugin should have methods hook to the legacy events and
    call appropriate methods here to transmit the "legacy action" to the xmpp
    user. This should look like this:

    .. code-block:python

        class Session(BaseSession):
            ...

            async def on_cool_chat_network_new_text_message(self, legacy_msg_event):
                contact = self.contacts.by_legacy_id(legacy_msg_event.from)
                contact.send_text(legacy_msg_event.text)

            async def on_cool_chat_network_new_typing_event(self, legacy_typing_event):
                contact = self.contacts.by_legacy_id(legacy_msg_event.from)
                contact.composing()
            ...
    """

    RESOURCE: str = "slidge"
    """
    A full JID, including a resource part is required for chat states (and maybe other stuff)
    to work properly. This is the name of the resource the contacts will use.
    """

    AVATAR = True
    RECEIPTS = True
    MARKS = True
    CHAT_STATES = True
    UPLOAD = True
    CORRECTION = True
    REACTION = True
    RETRACTION = True
    REPLIES = True

    """
    A list of features advertised through service discovery and client capabilities.
    """

    def __init__(
        self,
        session: "SessionType",
        legacy_id: LegacyContactIdType,
        jid_username: str,
    ):
        """
        :param session: The session this contact is part of
        :param legacy_id: The contact's legacy ID
        :param jid_username: User part of this contact's 'puppet' JID.
            NB: case-insensitive, and some special characters are not allowed
        """
        self.session = session
        self.user = session.user
        self.legacy_id = legacy_id
        self.jid_username = jid_username

        self.added_to_roster = False

        self._name: Optional[str] = None
        self._avatar: Optional[AvatarType] = None

        self.xmpp = session.xmpp
        self.xmpp.loop.create_task(self.__make_caps())

    def __repr__(self):
        return f"<LegacyContact <{self.jid}> ('{self.legacy_id}') of <{self.user}>"

    async def __make_caps(self):
        """
        Configure slixmpp to correctly advertise this contact's capabilities.
        """
        jid = self.jid
        xmpp = self.xmpp

        xmpp["xep_0030"].add_identity(jid=jid, category="client", itype="bot")
        add_feature = functools.partial(xmpp["xep_0030"].add_feature, jid=jid)
        if self.CHAT_STATES:
            await add_feature("http://jabber.org/protocol/chatstates")
        if self.AVATAR:
            await add_feature("vcard-temp")
        if self.RECEIPTS:
            await add_feature("urn:xmpp:receipts")
        if self.CORRECTION:
            await add_feature("urn:xmpp:message-correct:0")
        if self.MARKS:
            await add_feature("urn:xmpp:chat-markers:0")
        if self.UPLOAD:
            await add_feature("jabber:x:oob")
        if self.REACTION:
            await add_feature("urn:xmpp:reactions:0")
        if self.RETRACTION:
            await add_feature("urn:xmpp:message-retract:0")
        if self.REPLIES:
            await add_feature("urn:xmpp:reply:0")

        info = await xmpp["xep_0030"].get_info(jid, node=None, local=True)
        if isinstance(info, Iq):
            info = info["disco_info"]
        ver = xmpp["xep_0115"].generate_verstring(info, xmpp["xep_0115"].hash)
        await xmpp["xep_0030"].set_info(
            jid=jid,
            node="%s#%s" % (xmpp["xep_0115"].caps_node, ver),
            info=info,
        )

        await xmpp["xep_0115"].cache_caps(ver, info)
        await xmpp["xep_0115"].assign_verstring(jid, ver)

    async def __make_vcard(self):
        """
        Configure slixmpp to correctly set this contact's vcard (in fact only its avatar ATM)
        """
        await self.xmpp.set_vcard_avatar(jid=self.jid, avatar=self.avatar)

    @property
    def jid(self) -> JID:
        """
        Full JID (including the 'puppet' resource) of the contact
        """
        j = JID(self.jid_username + "@" + self.xmpp.boundjid.bare)
        j.resource = self.RESOURCE
        return j

    @property
    def name(self):
        """
        Friendly name of the contact, as it should appear in the user's roster
        """
        return self._name

    @name.setter
    def name(self, n: Optional[str]):
        self._name = n

    @property
    def avatar(self):
        """
        An image that represents this contact
        """
        return self._avatar

    @avatar.setter
    def avatar(self, a: Optional[AvatarType]):
        self._avatar = a
        self.xmpp.loop.create_task(self.__make_vcard())

    async def add_to_roster(self):
        """
        Add this contact to the user roster using :xep:`0356`
        """
        if self.xmpp.no_roster_push:
            log.debug("Roster push request by plugin ignored (--no-roster-push)")
            return
        kw = dict(
            jid=self.user.jid,
            roster_items={
                self.jid.bare: {
                    "name": self.name,
                    "subscription": "both",
                    "groups": [self.xmpp.ROSTER_GROUP],
                }
            },
        )
        try:
            await self.xmpp["xep_0356"].set_roster(**kw)
        except PermissionError:
            await self.xmpp["xep_0356_old"].set_roster(**kw)

        self.added_to_roster = True

    def online(self, status: Optional[str] = None):
        """
        Send an "online" presence from this contact to the user.

        :param status: Arbitrary text, details of the status, eg: "Listening to Britney Spears"
        """
        self.xmpp.send_presence(pfrom=self.jid, pto=self.user.jid.bare, pstatus=status)

    def away(self, status: Optional[str] = None):
        """
        Send an "away" presence from this contact to the user.

        This is a global status, as opposed to :meth:`.LegacyContact.inactive`
        which concerns a specific conversation, ie a specific "chat window"

        :param status: Arbitrary text, details of the status, eg: "Gone to fight capitalism"
        """
        self.xmpp.send_presence(
            pfrom=self.jid, pto=self.user.jid.bare, pshow="away", pstatus=status
        )

    def busy(self, status: Optional[str] = None):
        """
        Send a "busy" presence from this contact to the user,

        :param status: eg: "Trying to make sense of XEP-0100"
        """
        self.xmpp.send_presence(
            pfrom=self.jid, pto=self.user.jid.bare, pshow="busy", pstatus=status
        )

    def offline(self):
        """
        Send an "offline" presence from this contact to the user.
        """
        self.xmpp.send_presence(
            pfrom=self.jid, pto=self.user.jid.bare, ptype="unavailable"
        )

    def unsubscribe(self):
        """
        Send an "unsubscribed" presence from this contact to the user.
        """
        self.xmpp.send_presence(
            pfrom=self.jid, pto=self.user.jid.bare, ptype="unsubscribed"
        )

    def status(self, text: str):
        """
        Set a contact's status
        """
        self.xmpp.send_presence(pfrom=self.jid, pto=self.user.jid.bare, pstatus=text)

    def __chat_state(self, state: str):
        msg = self.xmpp.make_message(mfrom=self.jid, mto=self.user.jid, mtype="chat")
        msg["chat_state"] = state
        msg.enable("no-store")
        msg.send()

    def active(self):
        """
        Send an "active" chat state (:xep:`0085`) from this contact to the user.
        """
        self.__chat_state("active")

    def composing(self):
        """
        Send a "composing" (ie "typing notification") chat state (:xep:`0085`) from this contact to the user.
        """
        self.__chat_state("composing")

    def paused(self):
        """
        Send a "paused" (ie "typing paused notification") chat state (:xep:`0085`) from this contact to the user.
        """
        self.__chat_state("paused")

    def inactive(self):
        """
        Send an "inactive" (ie "typing paused notification") chat state (:xep:`0085`) from this contact to the user.
        """
        log.debug("%s go inactive", self)
        self.__chat_state("inactive")

    def __send_marker(
        self,
        legacy_msg_id: LegacyMessageType,
        marker: Literal["acknowledged", "received", "displayed"],
    ):
        """
        Send a message marker (:xep:`0333`) from this contact to the user.

        NB: for the 'received' marker, this also sends a message receipt (:xep:`0184`)

        :param legacy_msg_id: ID of the message this marker refers to
        :param marker: The marker type

        """
        xmpp_id = self.session.sent.get(legacy_msg_id)
        if xmpp_id is None:
            log.debug("Cannot find the XMPP ID of this msg: %s", legacy_msg_id)
        else:
            if marker == "received":
                receipt = self.xmpp.Message()
                receipt["to"] = self.user.jid
                receipt["receipt"] = xmpp_id
                receipt["from"] = self.jid
                receipt.send()
            self.xmpp["xep_0333"].send_marker(
                mto=self.user.jid,
                id=xmpp_id,
                marker=marker,
                mfrom=self.jid,
            )

    def ack(self, legacy_msg_id: LegacyMessageType):
        """
        Send an "acknowledged" message marker (:xep:`0333`) from this contact to the user.

        :param legacy_msg_id: The message this marker refers to
        """
        self.__send_marker(legacy_msg_id, "acknowledged")

    def received(self, legacy_msg_id: LegacyMessageType):
        """
        Send a "received" message marker (:xep:`0333`) and a "message delivery receipt"
        (:xep:`0184`)
        from this contact to the user

        :param legacy_msg_id: The message this marker refers to
        """
        self.__send_marker(legacy_msg_id, "received")

    def displayed(self, legacy_msg_id: LegacyMessageType):
        """
        Send a "displayed" message marker (:xep:`0333`) from this contact to the user.

        :param legacy_msg_id: The message this marker refers to
        """
        self.__send_marker(legacy_msg_id, "displayed")

    def __make_message(self, **kwargs) -> Message:
        m = self.xmpp.make_message(mfrom=self.jid, mto=self.user.jid, **kwargs)
        m.enable("markable")
        return m

    def __send_message(self, msg: Message, legacy_msg_id: Optional[Any] = None):
        if legacy_msg_id is not None:
            msg.set_id(self.session.legacy_msg_id_to_xmpp_msg_id(legacy_msg_id))
        msg.send()

    def __make_reply(self, msg: Message, reply_to_msg_id: Optional[LegacyMessageType]):
        if reply_to_msg_id is None:
            return
        xmpp_id = self.session.sent.get(
            reply_to_msg_id
        ) or self.session.legacy_msg_id_to_xmpp_msg_id(reply_to_msg_id)
        msg["reply"]["id"] = self.session.legacy_msg_id_to_xmpp_msg_id(xmpp_id)
        # FIXME: https://xmpp.org/extensions/xep-0461.html#usecases mentions that a full JID must be used here
        msg["reply"]["to"] = self.user.jid

    def send_text(
        self,
        body: str = "",
        *,
        chat_state: Optional[str] = "active",
        legacy_msg_id: Optional[LegacyMessageType] = None,
        reply_to_msg_id: Optional[LegacyMessageType] = None,
    ) -> Message:
        """
        Transmit a message from the contact to the user

        :param body: Context of the message
        :param chat_state: By default, will send an "active" chat state (:xep:`0085`) along with the
            message. Set this to ``None`` if this is not desired.
        :param legacy_msg_id: If you want to be able to transport read markers from the gateway
            user to the legacy network, specify this
        :param reply_to_msg_id:

        :return: the XMPP message that was sent
        """
        msg = self.__make_message(mbody=body, mtype="chat")
        if self.CHAT_STATES and chat_state is not None:
            msg["chat_state"] = chat_state
        self.__make_reply(msg, reply_to_msg_id)
        self.__send_message(msg, legacy_msg_id)
        return msg

    async def send_file(
        self,
        filename: Union[Path, str],
        content_type: Optional[str] = None,
        input_file: Optional[IO[bytes]] = None,
        url: Optional[str] = None,
        *,
        legacy_msg_id: Optional[LegacyMessageType] = None,
        reply_to_msg_id: Optional[LegacyMessageType] = None,
    ) -> Message:
        """
        Send a file using HTTP upload (:xep:`0363`)

        :param filename: Filename to use or location on disk to the file to upload
        :param content_type: MIME type, inferred from filename if not given
        :param input_file: Optionally, a file like object instead of a file on disk.
            filename will still be used to give the uploaded file a name
        :param legacy_msg_id: If you want to be able to transport read markers from the gateway
            user to the legacy network, specify this
        :param url: Optionally, a URL of a file that slidge will download and upload to the
            default file upload service on the xmpp server it's running on. url and input_file
            are mutually exclusive.
        :param reply_to_msg_id:

        :return: The sent msg stanza
        """
        log.debug("HOST: %s", self.xmpp.server_host)
        if url is not None:
            if input_file is not None:
                raise TypeError("Either a URL or a file-like object")
            async with aiohttp.ClientSession() as session:
                async with session.get(url) as r:
                    input_file = BytesIO(await r.read())
        url = await self.xmpp["xep_0363"].upload_file(
            filename=filename,
            content_type=content_type,
            input_file=input_file,
        )
        msg = self.__make_message()
        self.__make_reply(msg, reply_to_msg_id)
        msg["oob"]["url"] = url
        msg["body"] = url
        self.__send_message(msg, legacy_msg_id)
        return msg

    def __privileged_send(self, msg: Message):
        msg.set_from(self.user.jid.bare)
        msg.enable("store")
        self.session.ignore_messages.add(msg.get_id())
        try:
            self.xmpp["xep_0356"].send_privileged_message(msg)
        except PermissionError:
            self.xmpp["xep_0356_old"].send_privileged_message(msg)
        return msg.get_id()

    def carbon(
        self,
        body: str,
        legacy_id: Optional[Any] = None,
        date: Optional[datetime] = None,
    ):
        """
        Call this when the user sends a message to a legacy network contact.

        This synchronizes the outgoing message history on the XMPP side, using
         xep:`0356` to impersonate the XMPP user and send a message from the user to
        the contact. Thw XMPP server should in turn send carbons (:xep:`0280`) to online
        XMPP clients +/- write the message in server-side archives (:xep:`0_313`),
        depending on the user's and the server's archiving policy.

        :param str body: Body of the message.
        :param legacy_id: Legacy message ID
        :param str date: When was this message sent.
        """
        # we use Message() directly because we need xmlns="jabber:client"
        msg = Message()
        msg["to"] = self.jid.bare
        msg["type"] = "chat"
        msg["body"] = body
        if legacy_id:
            xmpp_id = self.session.legacy_msg_id_to_xmpp_msg_id(legacy_id)
            msg.set_id(xmpp_id)
            self.session.sent[legacy_id] = xmpp_id
        if date:
            if date.tzinfo is None:
                date = date.astimezone(timezone.utc)
            msg["delay"].set_stamp(date)

        return self.__privileged_send(msg)

    def carbon_read(self, legacy_msg_id: Any, date: Optional[datetime] = None):
        """
        Synchronize user read state from official clients.

        :param str legacy_msg_id:
        :param str date:
        """
        # we use Message() directly because we need xmlns="jabber:client"
        msg = Message()
        msg["to"] = self.jid.bare
        msg["type"] = "chat"
        msg["displayed"]["id"] = self.session.legacy_msg_id_to_xmpp_msg_id(
            legacy_msg_id
        )
        if date is not None:
            if date.tzinfo is None:
                date = date.astimezone(timezone.utc)
            msg["delay"].set_stamp(date)

        return self.__privileged_send(msg)

    def carbon_correct(self, legacy_msg_id: LegacyMessageType, text: str):
        """
        Call this when the user corrects their own (last) message from an official client

        :param legacy_msg_id:
        :param text: The new body of the message
        """
        if (xmpp_id := self.session.sent.get(legacy_msg_id)) is None:
            log.debug(
                "Cannot find XMPP ID of msg '%s' corrected from the official client",
                legacy_msg_id,
            )
            return
        msg = Message()
        msg.set_to(self.jid.bare)
        msg.set_type("chat")
        msg["replace"]["id"] = xmpp_id
        msg["body"] = text
        return self.__privileged_send(msg)

    def carbon_react(
        self, legacy_msg_id: LegacyMessageType, reactions: Iterable[str] = ()
    ):
        """
        Call this to modify the user's own reactions (:xep:`0444`) about a message.

        Can be called when the user reacts from the official client, or to modify a user's
        reaction when the legacy network has constraints about acceptable reactions.

        :param legacy_msg_id: Legacy message ID this refers to
        :param reactions: iterable of emojis
        """
        msg = Message()
        msg["to"] = self.jid.bare
        msg["type"] = "chat"
        self.xmpp["xep_0444"].set_reactions(
            msg,
            to_id=self.session.legacy_msg_id_to_xmpp_msg_id(legacy_msg_id),
            reactions=reactions,
        )
        return self.__privileged_send(msg)

    def carbon_retract(self, legacy_msg_id):
        if (xmpp_id := self.session.sent.inverse.get(legacy_msg_id)) is None:
            if (xmpp_id := self.session.sent.get(legacy_msg_id)) is None:
                log.debug("Cannot find XMPP ID of retracted msg: %s", legacy_msg_id)
                return

        msg = Message()
        msg.set_to(self.jid.bare)
        msg.set_type("chat")
        msg["apply_to"]["id"] = xmpp_id
        msg["apply_to"].enable("retract")
        return self.__privileged_send(msg)

    def correct(self, legacy_msg_id: Any, new_text: str):
        """
        Call this when a legacy contact has modified his last message content.

        Uses last message correction (:xep:`0308`)

        :param legacy_msg_id: Legacy message ID this correction refers to
        :param new_text: The new text
        """
        msg = self.__make_message()
        msg["replace"]["id"] = self.session.legacy_msg_id_to_xmpp_msg_id(legacy_msg_id)
        msg["body"] = new_text
        self.__send_message(msg)

    def react(self, legacy_msg_id: LegacyMessageType, emojis: Iterable[str]):
        """
        Call this when a legacy contact reacts to a message

        :param legacy_msg_id: The message which the reaction refers to.
        :param emojis: A iterable of emojis used as reactions
        :return:
        """
        if (xmpp_id := self.session.sent.get(legacy_msg_id)) is None:
            log.debug(
                "Cannot determine which message this reaction refers to, attempting msg ID conversion"
            )
            xmpp_id = self.session.legacy_msg_id_to_xmpp_msg_id(legacy_msg_id)
        msg = self.__make_message()
        self.xmpp["xep_0444"].set_reactions(
            msg,
            to_id=xmpp_id,
            reactions=emojis,
        )
        self.__send_message(msg)
        return msg

    def retract(self, legacy_msg_id: LegacyMessageType):
        """
        Call this when a legacy contact retracts (:XEP:`0424`) a message

        :param legacy_msg_id: Legacy ID of the message to delete
        """
        self.xmpp["xep_0424"].send_retraction(
            mto=self.user.jid,
            mfrom=self.jid,
            include_fallback=True,
            fallback_text="I have deleted the message %s, but your XMPP client does not support that"
            % legacy_msg_id,  # https://github.com/movim/movim/issues/1074
            id=self.session.legacy_msg_id_to_xmpp_msg_id(legacy_msg_id),
        )


LegacyContactType = TypeVar("LegacyContactType", bound=LegacyContact)


class LegacyRoster(Generic[LegacyContactType, SessionType], metaclass=SubclassableOnce):
    """
    Virtual roster of a gateway user, that allows to represent all
    of their contacts as singleton instances (if used properly and not too bugged).

    Every :class:`.BaseSession` instance will have its own :class:`.LegacyRoster` instance
    accessible via the :attr:`.BaseSession.contacts` attribute.

    Typically, you will mostly use the :meth:`.LegacyRoster.by_legacy_id` function to
    retrieve a contact instance.

    You might need to override :meth:`.LegacyRoster.legacy_id_to_jid_username` and/or
    :meth:`.LegacyRoster.jid_username_to_legacy_id` to incorporate some custom logic
    if you need some characters when translation JID user parts and legacy IDs.
    """

    def __init__(self, session: "SessionType"):
        self._contact_cls: Type[
            LegacyContactType
        ] = LegacyContact.get_self_or_unique_subclass()
        self._contact_cls.xmpp = session.xmpp

        self.session = session
        self._contacts_by_bare_jid: dict[str, LegacyContactType] = {}
        self._contacts_by_legacy_id: dict[LegacyContactIdType, LegacyContactType] = {}

    def __iter__(self):
        return iter(self._contacts_by_legacy_id.values())

    def by_jid(self, contact_jid: JID) -> LegacyContactType:
        """
        Retrieve a contact by their JID

        If the contact was not instantiated before, it will be created
        using :meth:`slidge.LegacyRoster.jid_username_to_legacy_id` to infer their
        legacy user ID.

        :param contact_jid:
        :return:
        """
        bare = contact_jid.bare
        c = self._contacts_by_bare_jid.get(bare)
        if c is None:
            jid_username = str(contact_jid.username)
            log.debug("Contact %s not found", contact_jid)
            c = self._contact_cls(
                self.session,
                self.jid_username_to_legacy_id(jid_username),
                jid_username,
            )
            self._contacts_by_bare_jid[bare] = c
        return c

    def by_legacy_id(self, legacy_id: Any) -> LegacyContactType:
        """
        Retrieve a contact by their legacy_id

        If the contact was not instantiated before, it will be created
        using :meth:`slidge.LegacyRoster.legacy_id_to_jid_username` to infer their
        legacy user ID.

        :param legacy_id:
        :return:
        """
        c = self._contacts_by_legacy_id.get(legacy_id)
        if c is None:
            log.debug("Contact %s not found in roster", legacy_id)
            c = self._contact_cls(
                self.session, legacy_id, self.legacy_id_to_jid_username(legacy_id)
            )
            self._contacts_by_legacy_id[legacy_id] = c
        return c

    def by_stanza(self, s) -> LegacyContactType:
        """
        Retrieve a contact by the destination of a stanza

        See :meth:`slidge.Roster.by_legacy_id` for more info.

        :param s:
        :return:
        """
        return self.by_jid(s.get_to())

    @staticmethod
    def legacy_id_to_jid_username(legacy_id: Any) -> str:
        """
        Convert a legacy ID to a valid 'user' part of a JID

        Should be overridden for cases where the str conversion of
        the legacy_id is not enough, e.g., if it contains forbidden character.

        :param legacy_id:
        """
        return str(legacy_id)

    @staticmethod
    def jid_username_to_legacy_id(jid_username: str) -> LegacyUserIdType:
        """
        Convert a JID user part to a legacy ID.

        Should be overridden in case legacy IDs are not strings, or more generally
        for any case where the username part of a JID is not enough to identify
        a contact on the legacy network.

        Default implementation is an identity operation

        :param jid_username: User part of a JID, ie "user" in "user@example.com"
        :return: An identifier for the user on the legacy network.
        """
        return jid_username  # type:ignore


LegacyRosterType = TypeVar("LegacyRosterType", bound=LegacyRoster)

log = logging.getLogger(__name__)



import asyncio
import functools
from typing import TYPE_CHECKING, Any, Union

import discord as di
from slixmpp.exceptions import XMPPError

from slidge import *

if TYPE_CHECKING:
    from . import Contact, Gateway, Roster
    from .client import Discord


def raise_xmpp_not_found_if_necessary(func):
    @functools.wraps(func)
    def wrapped(*a, **kw):
        contact: "Contact" = a[-1]
        if contact.discord_id is None:
            raise XMPPError("not-found")
        return func(*a, **kw)

    return wrapped


class Session(BaseSession["Contact", "Roster", "Gateway"]):
    discord: "Discord"
    ready_future: asyncio.Future[bool]
    delete_futures: dict[int, asyncio.Future[bool]]
    edit_futures: dict[int, asyncio.Future[bool]]
    send_futures: dict[int, asyncio.Future[bool]]
    send_lock: asyncio.Lock

    def post_init(self):
        from .client import Discord

        self.discord = Discord(self)
        self.ready_future = self.xmpp.loop.create_future()
        self.delete_futures = {}
        self.edit_futures = {}
        self.send_futures = {}
        self.send_lock = asyncio.Lock()

    @staticmethod
    def xmpp_msg_id_to_legacy_msg_id(i: str) -> Union[int, str]:
        try:
            return int(i)
        except ValueError:
            return i

    async def login(self):
        self.xmpp.loop.create_task(
            self.discord.start(self.user.registration_form["token"])
        )
        await self.ready_future
        for u in self.discord.users:
            if not isinstance(u, di.User):
                self.log.debug(f"Skipping %s", u)
                continue
            if not u.is_friend():
                self.log.debug(f"%s is not a friend", u)
                continue
            c = self.contacts.by_legacy_id(str(u))
            c.name = u.display_name
            c.avatar = str(u.avatar_url)
            self.log.debug("Avatar: %s", u.avatar_url)
            c.discord_id = u.id
            await c.add_to_roster()
            c.online()
        return f"Logged on as {self.discord.user}"

    @raise_xmpp_not_found_if_necessary
    async def send_text(self, t: str, c: "Contact", *, reply_to_msg_id=None):
        async with self.send_lock:
            mid = (
                await self.discord.get_user(c.discord_id).send(
                    t,
                    reference=None
                    if reply_to_msg_id is None
                    else di.MessageReference(
                        message_id=reply_to_msg_id,
                        channel_id=c.direct_channel_id,
                    ),
                )
            ).id
        f = self.send_futures[mid] = self.xmpp.loop.create_future()
        await f
        return mid

    async def logout(self):
        await self.discord.close()

    async def send_file(self, u: str, c: "Contact", *, reply_to_msg_id=None):
        # discord clients inline previews of external URLs, so no need to actually send on discord servers
        await self.discord.get_user(c.discord_id).send(u)

    async def active(self, c: "Contact"):
        pass

    async def inactive(self, c: "Contact"):
        pass

    @raise_xmpp_not_found_if_necessary
    async def composing(self, c: "Contact"):
        await self.discord.get_user(c.discord_id).trigger_typing()

    async def paused(self, c: "Contact"):
        pass

    @raise_xmpp_not_found_if_necessary
    async def displayed(self, legacy_msg_id: str, c: "Contact"):
        if not isinstance(legacy_msg_id, int):
            self.log.debug("This is not a valid discord msg id: %s", legacy_msg_id)
            return
        u: di.User = self.discord.get_user(c.discord_id)
        channel: di.DMChannel = u.dm_channel
        if channel is None:
            return
        m = await channel.fetch_message(legacy_msg_id)
        self.log.debug("Message %s should be marked as read", m)
        # try:
        #     await m.ack()  # triggers 404, maybe does not work for DM?
        # except Exception as e:
        #     self.log.exception("Message %s should have been marked as read but this raised %s", m, e)

    @raise_xmpp_not_found_if_necessary
    async def correct(self, text: str, legacy_msg_id: Any, c: "Contact"):
        u: di.User = self.discord.get_user(c.discord_id)
        channel: di.DMChannel = u.dm_channel
        if channel is None:
            return
        m = await channel.fetch_message(legacy_msg_id)
        self.edit_futures[legacy_msg_id] = self.xmpp.loop.create_future()
        await m.edit(content=text)
        await self.edit_futures[legacy_msg_id]

    @raise_xmpp_not_found_if_necessary
    async def react(self, legacy_msg_id: int, emojis: list[str], c: "Contact"):
        u: di.User = self.discord.get_user(c.discord_id)
        channel: di.DMChannel = u.dm_channel
        if channel is None:
            return
        m = await channel.fetch_message(legacy_msg_id)

        legacy_reactions = set(self.get_my_legacy_reactions(m))
        xmpp_reactions = set(emojis)

        self.log.debug("%s vs %s", legacy_reactions, xmpp_reactions)
        for e in xmpp_reactions - legacy_reactions:
            await m.add_reaction(e)
        for e in legacy_reactions - xmpp_reactions:
            await m.remove_reaction(e, self.discord.user)

    @raise_xmpp_not_found_if_necessary
    async def retract(self, legacy_msg_id: Any, c: "Contact"):
        u: di.User = self.discord.get_user(c.discord_id)
        channel: di.DMChannel = u.dm_channel
        if channel is None:
            return
        m = await channel.fetch_message(legacy_msg_id)
        self.delete_futures[legacy_msg_id] = self.xmpp.loop.create_future()
        await m.delete()
        await self.delete_futures[legacy_msg_id]

    def update_reactions(self, message: di.Message):
        self.contacts.by_discord_user(message.channel.recipient).carbon_react(
            message.id, self.get_my_legacy_reactions(message)
        )

    @staticmethod
    def get_my_legacy_reactions(message: di.Message) -> list[str]:
        reactions = []
        for r in message.reactions:
            if r.me and not r.custom_emoji:
                reactions.append(r.emoji)

        return reactions

    async def search(self, form_values: dict[str, str]) -> SearchResult:
        pass



import asyncio
import io
import pprint
import re
from datetime import datetime
from typing import Any, Optional

import emoji
from mattermost_api_reference_client.models import Status
from mattermost_api_reference_client.types import Unset

from slidge import *

from .api import MattermostClient
from .websocket import EventType, MattermostEvent, Websocket


class Gateway(BaseGateway):
    REGISTRATION_INSTRUCTIONS = (
        "Enter mattermost credentials. "
        "Get your MMAUTH_TOKEN on the web interface, using the dev tools of your browser (it's a cookie)."
    )
    REGISTRATION_FIELDS = [
        FormField(var="url", label="Mattermost server URL", required=True),
        FormField(var="token", label="MMAUTH_TOKEN", required=True),
        FormField(var="basepath", label="Base path", value="/api/v4", required=True),
        FormField(
            var="basepath_ws",
            label="Websocket base path",
            value="/websocket",
            required=True,
        ),
        FormField(
            var="strict_ssl",
            label="Strict SSL verification",
            value="1",
            required=False,
            type="boolean",
        ),
    ]

    ROSTER_GROUP = "Mattermost"

    COMPONENT_NAME = "Mattermost (slidge)"
    COMPONENT_TYPE = "mattermost"

    COMPONENT_AVATAR = "https://play-lh.googleusercontent.com/aX7JaAPkmnkeThK4kgb_HHlBnswXF0sPyNI8I8LNmEMMo1vDvMx32tCzgPMsyEXXzZRc"


class Contact(LegacyContact["Session"]):
    legacy_id: str

    MARKS = False

    def __init__(self, *a, **kw):
        super().__init__(*a, **kw)
        self._direct_channel_id: Optional[str] = None
        self._mm_id: Optional[str] = None

    def update_status(self, status: Optional[str]):
        if status is None:  # custom status
            self.session.log.debug("Status is None: %s", status)
            self.online()
        elif status == "online":
            self.online()
        elif status == "offline":
            self.offline()
        elif status == "away":
            self.away()
        elif status == "dnd":
            self.busy()
        else:
            self.session.log.warning(
                "Unknown status for '%s':",
                status,
            )

    async def direct_channel_id(self):
        if self._direct_channel_id is None:
            self._direct_channel_id = (
                await self.session.mm_client.get_direct_channel(await self.mm_id())
            ).id
            self.session.contacts.direct_channel_id_to_username[
                self._direct_channel_id
            ] = self.legacy_id
        return self._direct_channel_id

    async def mm_id(self):
        if self._mm_id is None:
            self._mm_id = (
                await self.session.mm_client.get_user_by_username(self.legacy_id)
            ).id
            self.session.contacts.user_id_to_username[self._mm_id] = self.legacy_id
        return self._mm_id

    async def update_reactions(self, legacy_msg_id):
        self.react(
            legacy_msg_id,
            [
                emoji.emojize(f":{x}:", language="alias")
                for x in await self.session.get_mm_reactions(
                    legacy_msg_id, await self.mm_id()
                )
            ],
        )


class Roster(LegacyRoster[Contact, "Session"]):
    user_id_to_username: dict[str, str]
    direct_channel_id_to_username: dict[str, str]

    def __init__(self, *a, **kw):
        super().__init__(*a, **kw)
        self.user_id_to_username = {}
        self.direct_channel_id_to_username = {}

    async def by_mm_user_id(self, user_id: str):
        try:
            legacy_id = self.user_id_to_username[user_id]
        except KeyError:
            user = await self.session.mm_client.get_user(user_id)
            if isinstance(user.username, Unset):
                raise RuntimeError
            legacy_id = self.user_id_to_username[user_id] = user.username
        return self.by_legacy_id(legacy_id)

    async def by_direct_channel_id(self, channel_id: str):
        if (username := self.direct_channel_id_to_username.get(channel_id)) is None:
            for c in self:
                if (await c.direct_channel_id()) == channel_id:
                    return c
        else:
            return self.by_legacy_id(username)


class Session(BaseSession[Contact, Roster, Gateway]):
    mm_client: MattermostClient
    ws: Websocket
    messages_waiting_for_echo: set[str]
    send_lock: asyncio.Lock

    def post_init(self):
        self.messages_waiting_for_echo = set()
        self.send_lock = asyncio.Lock()
        f = self.user.registration_form
        url = f["url"] + f["basepath"]
        self.mm_client = MattermostClient(
            url,
            verify_ssl=f["strict_ssl"],
            timeout=5,
            token=f["token"],
        )
        self.ws = Websocket(
            re.sub("^http", "ws", f["url"]) + f["basepath"] + f["basepath_ws"],
            f["token"],
        )

    async def login(self):
        await self.mm_client.login()

        await self.add_contacts()
        self.xmpp.loop.create_task(self.ws.connect(self.on_mm_event))
        if self.mm_client.me is None:
            raise RuntimeError

        return f"Connected as '{(await self.mm_client.me).username}'"

    async def add_contacts(self):
        user_ids = await self.mm_client.get_contacts()
        contact_mm_users = await self.mm_client.get_users_by_ids(user_ids)
        contact_mm_statuses = await self.mm_client.get_users_statuses_by_ids(user_ids)

        statuses = {s.user_id: s for s in contact_mm_statuses}

        for user in contact_mm_users:
            status: Status = statuses[user.id]
            contact = self.contacts.by_legacy_id(user.username)
            self.contacts.user_id_to_username[user.id] = user.username
            if user.nickname:
                contact.name = user.nickname
            elif user.first_name and user.last_name:
                contact.name = user.first_name + " " + user.last_name
            elif user.first_name:
                contact.name = user.first_name
            elif user.last_name:
                contact.name = user.last_name

            contact.avatar = await self.mm_client.get_profile_image(user.id)

            await contact.add_to_roster()
            contact.update_status(status.status)

    async def on_mm_event(self, event: MattermostEvent):
        self.log.debug("Event: %s", event)
        if event.type == EventType.Hello:
            self.log.info("Received hello event: %s", event.data)
        elif event.type == EventType.Posted:
            post = event.data["post"]
            self.log.debug("Post: %s", pprint.pformat(post))

            message = post["message"]

            channel_id = post["channel_id"]
            post_id = post["id"]
            user_id = post["user_id"]

            if event.data["channel_type"] == "D":  # Direct messages?
                if user_id == await self.mm_client.mm_id:
                    try:
                        async with self.send_lock:
                            self.messages_waiting_for_echo.remove(post_id)
                    except KeyError:
                        members = await self.mm_client.get_channel_members(channel_id)
                        if len(members) > 2:
                            raise RuntimeError("Not a direct message after all")
                        for m in members:
                            if m.user_id != await self.mm_client.mm_id:
                                contact = await self.contacts.by_mm_user_id(m.user_id)
                                break
                        else:
                            raise RuntimeError("What?")

                        contact.carbon(
                            message,
                            post_id,
                            datetime.fromtimestamp(post["update_at"] / 1000),
                        )
                else:
                    contact = await self.contacts.by_mm_user_id(user_id)
                    if event.data.get("set_online"):
                        contact.online()
                    contact.send_text(message, legacy_msg_id=post_id)
                    for file_meta in post.get("metadata", {}).get("files", []):
                        await contact.send_file(
                            filename=file_meta["name"],
                            input_file=io.BytesIO(
                                await self.mm_client.get_file(file_meta["id"])
                            ),
                        )
            elif event.data["channel_type"] == "P":
                # private channel
                pass
        elif event.type == EventType.ChannelViewed:
            channel_id = event.data["channel_id"]
            posts = await self.mm_client.get_posts_for_channel(channel_id)
            last_msg_id = posts.posts.additional_keys[-1]
            if (c := await self.contacts.by_direct_channel_id(channel_id)) is None:
                self.log.debug("Ignoring unknown channel")
            else:
                c.carbon_read(last_msg_id)
        elif event.type == EventType.StatusChange:
            user_id = event.data["user_id"]
            if user_id == await self.mm_client.mm_id:
                self.log.debug("Own status change")
            else:

                contact = await self.contacts.by_mm_user_id(user_id)
                contact.update_status(event.data["status"])
        elif event.type == EventType.Typing:
            contact = await self.contacts.by_mm_user_id(event.data["user_id"])
            contact.composing()
        elif event.type == EventType.PostEdited:
            post = event.data["post"]
            contact = await self.contacts.by_mm_user_id(post["user_id"])
            if post["channel_id"] == await contact.direct_channel_id():
                contact.correct(post["id"], post["message"])
        elif event.type == EventType.PostDeleted:
            post = event.data["post"]
            contact = await self.contacts.by_mm_user_id(post["user_id"])
            if post["channel_id"] == await contact.direct_channel_id():
                contact.retract(post["id"])
        elif event.type in (EventType.ReactionAdded, EventType.ReactionRemoved):
            reaction = event.data["reaction"]
            if (who := reaction["user_id"]) == await self.mm_client.mm_id:
                pass
            else:
                await (await self.contacts.by_mm_user_id(who)).update_reactions(
                    reaction["post_id"]
                )

    async def logout(self):
        pass

    async def send_text(self, t: str, c: Contact, *, reply_to_msg_id=None):
        async with self.send_lock:
            msg_id = await self.mm_client.send_message_to_user(c.legacy_id, t)
            self.messages_waiting_for_echo.add(msg_id)
            return msg_id

    async def send_file(self, u: str, c: Contact, *, reply_to_msg_id=None):
        channel_id = await c.direct_channel_id()
        file_id = await self.mm_client.upload_file(channel_id, u)
        return await self.mm_client.send_message_with_file(channel_id, file_id)

    async def active(self, c: Contact):
        pass

    async def inactive(self, c: Contact):
        pass

    async def composing(self, c: Contact):
        await self.ws.user_typing(await c.direct_channel_id())

    async def paused(self, c: Contact):
        # no equivalent in MM, seems to have an automatic timeout in clients
        pass

    async def displayed(self, legacy_msg_id: Any, c: Contact):
        # no read marks in MM?
        pass

    async def correct(self, text: str, legacy_msg_id: Any, c: Contact):
        await self.mm_client.update_post(legacy_msg_id, text)

    async def search(self, form_values: dict[str, str]):
        pass

    async def retract(self, legacy_msg_id: Any, c: Contact):
        await self.mm_client.delete_post(legacy_msg_id)

    async def react(self, legacy_msg_id: Any, emojis: list[str], c: Contact):
        mm_reactions = await self.get_mm_reactions(
            legacy_msg_id, await self.mm_client.mm_id
        )
        xmpp_reactions = {
            emoji.demojize(x, language="alias", delimiters=("", "")) for x in emojis
        }
        self.log.debug("%s vs %s", mm_reactions, xmpp_reactions)
        for e in xmpp_reactions - mm_reactions:
            await self.mm_client.react(legacy_msg_id, e)
        for e in mm_reactions - xmpp_reactions:
            await self.mm_client.delete_reaction(legacy_msg_id, e)

    async def get_mm_reactions(self, legacy_msg_id: str, user_id: Optional[str]):
        return {
            # emoji.emojize(f":{x.emoji_name}:", language='alias')
            x.emoji_name
            for x in await self.mm_client.get_reactions(legacy_msg_id)
            if x.user_id == user_id
        }



import asyncio
import io
import logging

import aiohttp
import emoji
from mattermost_api_reference_client.api.channels import (
    create_direct_channel,
    get_channel_members,
    get_channels_for_team_for_user,
)
from mattermost_api_reference_client.api.files import get_file, upload_file
from mattermost_api_reference_client.api.posts import (
    create_post,
    delete_post,
    get_posts_for_channel,
    update_post,
)
from mattermost_api_reference_client.api.reactions import (
    delete_reaction,
    get_reactions,
    save_reaction,
)
from mattermost_api_reference_client.api.status import get_users_statuses_by_ids
from mattermost_api_reference_client.api.teams import get_teams_for_user
from mattermost_api_reference_client.api.users import (
    get_profile_image,
    get_user,
    get_user_by_username,
    get_users_by_ids,
)
from mattermost_api_reference_client.client import AuthenticatedClient
from mattermost_api_reference_client.models import Reaction, Status, User
from mattermost_api_reference_client.models.create_post_json_body import (
    CreatePostJsonBody,
)
from mattermost_api_reference_client.models.update_post_json_body import (
    UpdatePostJsonBody,
)
from mattermost_api_reference_client.models.upload_file_multipart_data import (
    UploadFileMultipartData,
)
from mattermost_api_reference_client.types import File, Unset


class MattermostClient:
    # TODO: this should be autogenerated using a template in mattermost_api_reference_client

    def __init__(self, *args, **kwargs):
        self.http = AuthenticatedClient(*args, **kwargs)
        self.mm_id: asyncio.Future[str] = asyncio.get_running_loop().create_future()
        self.me: asyncio.Future[User] = asyncio.get_running_loop().create_future()

    async def login(self):
        log.debug("Login")
        me = await get_user.asyncio("me", client=self.http)
        if me is None:
            raise RuntimeError("Could not login")
        self.me.set_result(me)
        my_id = me.id
        if isinstance(my_id, Unset):
            raise RuntimeError("Could not login")
        self.mm_id.set_result(my_id)
        log.debug("Me: %s", me)

    async def get_contacts(self) -> list[str]:
        mm = self.http
        my_id = await self.mm_id

        contact_mm_ids: list[str] = []

        teams = await get_teams_for_user.asyncio("me", client=mm)

        if teams is None:
            raise RuntimeError

        for team in teams:
            if isinstance(team.id, Unset):
                log.warning("Team without ID")
                continue
            channels = await get_channels_for_team_for_user.asyncio(
                "me", team.id, client=mm
            )

            if channels is None:
                log.warning("Team without channels")
                continue

            for channel in channels:
                if isinstance(channel.id, Unset):
                    log.warning("Channel without ID")
                    continue
                members = await self.get_channel_members(channel.id, per_page=4)
                if len(members) == 2:
                    user_ids = {m.user_id for m in members}
                    try:
                        user_ids.remove(my_id)
                    except KeyError:
                        log.warning("Weird 2 person channel: %s", members)
                    else:
                        contact_id = user_ids.pop()
                        if not isinstance(contact_id, str):
                            log.warning("Weird contact: %s", members)
                            continue
                        contact_mm_ids.append(contact_id)

        return contact_mm_ids

    async def get_channel_members(
        self, channel_id: str, *, page: int = 0, per_page: int = 10
    ):
        members = await get_channel_members.asyncio(
            channel_id, client=self.http, per_page=per_page, page=page
        )
        if members is None:
            raise RuntimeError
        return members

    async def get_users_by_ids(self, user_ids: list[str]) -> list[User]:
        r = await get_users_by_ids.asyncio(json_body=user_ids, client=self.http)
        if r is None:
            raise RuntimeError
        return r

    async def get_user(self, user_id: str):
        r = await get_user.asyncio(user_id, client=self.http)
        if r is None:
            raise RuntimeError
        if isinstance(r.username, Unset):
            raise RuntimeError
        return r

    async def get_users_statuses_by_ids(self, user_ids: list[str]) -> list[Status]:
        r = await get_users_statuses_by_ids.asyncio(
            json_body=user_ids, client=self.http
        )
        if r is None:
            raise RuntimeError
        return r

    async def send_message_to_user(self, user_id: str, text: str) -> str:
        await self.mm_id
        mm = self.http

        other = await self.get_user_by_username(user_id)

        direct_channel = await self.get_direct_channel(other.id)

        msg = await create_post.asyncio(
            json_body=CreatePostJsonBody(channel_id=direct_channel.id, message=text),
            client=mm,
        )
        if msg is None:
            raise RuntimeError

        if isinstance(msg.id, Unset):
            raise RuntimeError

        return msg.id

    async def send_message_with_file(self, channel_id: str, file_id: str):
        r = await create_post.asyncio(
            json_body=CreatePostJsonBody(
                channel_id=channel_id, file_ids=[file_id], message=""
            ),
            client=self.http,
        )
        if r is None or isinstance(r.id, Unset):
            raise RuntimeError(r)
        return r.id

    async def get_user_by_username(self, username: str) -> User:
        user = await get_user_by_username.asyncio(username, client=self.http)
        if user is None or isinstance(user.id, Unset):
            raise RuntimeError("Contact not found")
        return user

    async def get_direct_channel(self, user_id):
        direct_channel = await create_direct_channel.asyncio(
            json_body=[await self.mm_id, user_id], client=self.http
        )
        if direct_channel is None or isinstance(direct_channel.id, Unset):
            raise RuntimeError("Could not create direct channel")
        return direct_channel

    async def get_profile_image(self, user_id: str) -> bytes:
        resp = await get_profile_image.asyncio_detailed(user_id, client=self.http)
        return resp.content

    async def get_file(self, file_id: str):
        resp = await get_file.asyncio_detailed(file_id, client=self.http)
        return resp.content

    async def delete_post(self, post_id: str):
        r = await delete_post.asyncio(post_id, client=self.http)
        if r is not None and not isinstance(r, Unset) and r.status != "ok":
            raise RuntimeError("Could not delete post %s", post_id)

    async def update_post(self, post_id: str, body: str):
        r = await update_post.asyncio(
            post_id,
            client=self.http,
            json_body=UpdatePostJsonBody(id=post_id, message=body),
        )
        if r is None or isinstance(r, Unset):
            raise RuntimeError(r)
        return r.id

    async def get_posts_for_channel(self, channel_id: str):
        r = await get_posts_for_channel.asyncio(
            channel_id, client=self.http, per_page=2
        )
        if r is None or isinstance(r, Unset):
            raise RuntimeError(r)
        return r

    async def upload_file(self, channel_id: str, url: str):
        async with aiohttp.ClientSession() as s:
            async with s.get(url) as get_response:
                data = await get_response.read()
        req = UploadFileMultipartData(
            files=File(file_name=url.split("/")[-1], payload=io.BytesIO(data)),
            channel_id=channel_id,
        )
        r = await upload_file.asyncio(client=self.http, multipart_data=req)
        if (
            r is None
            or isinstance(r, Unset)
            or r.file_infos is None
            or isinstance(r.file_infos, Unset)
            or len(r.file_infos) != 1
        ):
            raise RuntimeError(r)
        return r.file_infos[0].id

    async def react(self, post_id: str, emoji_char: str):
        return await save_reaction.asyncio(
            client=self.http,
            json_body=Reaction(
                user_id=await self.mm_id,
                post_id=post_id,
                emoji_name=demojize(emoji_char),
            ),
        )

    async def get_reactions(self, post_id: str):
        try:
            return await get_reactions.asyncio(post_id, client=self.http)
        except TypeError:
            return []

    async def delete_reaction(self, post_id: str, emoji_name: str):
        await delete_reaction.asyncio(
            await self.mm_id, post_id, emoji_name=emoji_name, client=self.http
        )


def demojize(emoji_char: str):
    return emoji.demojize(emoji_char, delimiters=("", ""), language="alias")


log = logging.getLogger(__name__)



import asyncio
import functools
import logging
from argparse import ArgumentParser
from datetime import datetime
from typing import TYPE_CHECKING, Any, Optional

import aiosignald.exc as sigexc
import aiosignald.generated as sigapi
from aiosignald import SignaldAPI
from slixmpp import JID, Iq, Message
from slixmpp.exceptions import XMPPError
from slixmpp.plugins.xep_0004 import Form

from slidge import *

if TYPE_CHECKING:
    from .session import Session

from . import txt


class Gateway(BaseGateway):
    COMPONENT_NAME = "Signal (slidge)"
    COMPONENT_TYPE = "signal"
    COMPONENT_AVATAR = (
        "https://upload.wikimedia.org/wikipedia/commons/5/56/Logo_Signal..png"
    )
    REGISTRATION_INSTRUCTIONS = txt.REGISTRATION_INSTRUCTIONS
    REGISTRATION_FIELDS = txt.REGISTRATION_FIELDS

    ROSTER_GROUP = "Signal"

    signal: asyncio.Future["Signal"]
    signal_socket: str
    sessions_by_phone: dict[str, "Session"] = {}

    CHAT_COMMANDS = {
        "add_device": "_chat_command_add_device",
        "get_identities": "_chat_command_get_identities",
    }

    def __init__(self, args):
        super(Gateway, self).__init__(args)
        self.signal: asyncio.Future[Signal] = self.loop.create_future()

    def config(self, argv: list[str]):
        args = get_parser().parse_args(argv)
        self.signal_socket = socket = args.socket
        self.loop.create_task(self.connect_signal(socket))

    async def connect_signal(self, socket: str):
        """
        Establish connection to the signald socker
        """
        log.debug("Connecting to signald...")
        _, signal = await self.loop.create_unix_connection(
            functools.partial(Signal, self), socket
        )
        self.signal.set_result(signal)
        await signal.on_con_lost
        log.error("Signald UNIX socket connection lost!")
        raise RuntimeError("Signald socket connection lost")

    def add_adhoc_commands(self):
        self["xep_0050"].add_command(
            node="linked_devices",
            name="Get linked devices",
            handler=self._handle_linked_devices,
        )
        self["xep_0050"].add_command(
            node="add_device",
            name="Link a new device",
            handler=self._handle_add_device1,
        )

    async def _handle_linked_devices(self, iq: Iq, adhoc_session: dict[str, Any]):
        user = user_store.get_by_stanza(iq)
        if user is None:
            raise XMPPError("subscription-required")

        devices = await (await self.signal).get_linked_devices(
            account=user.registration_form["phone"]
        )

        # TODO: uncomment this when https://dev.gajim.org/gajim/gajim/-/issues/10857 is fixed
        # There are probably other clients that handle this just fine and this would make more sense
        # to use this, but I think targeting gajim compatibility when there are easy workarounds
        # is OK
        # form = self["xep_0004"].make_form("result", "Linked devices")
        # form.add_reported("id", label="ID", type="fixed")
        # form.add_reported("name", label="Name", type="fixed")
        # form.add_reported("created", label="Created", type="fixed")
        # form.add_reported("last_seen", label="Last seen", type="fixed")
        # for d in devices.devices:
        #     form.add_item(
        #         {
        #             "name": d.name,
        #             "id": str(d.id),
        #             "created": datetime.fromtimestamp(d.created / 1000).isoformat(),
        #             "last_seen": datetime.fromtimestamp(d.lastSeen / 1000).isoformat(),
        #         }
        #     )
        #
        # adhoc_session["payload"] = form
        adhoc_session["notes"] = [
            (
                "info",
                f"Name: {d.name} / "
                f"ID: {d.id} / "
                f"Created: {datetime.fromtimestamp(d.created / 1000).isoformat()} / "
                f"Last seen: {datetime.fromtimestamp(d.lastSeen / 1000).isoformat()}",
            )
            for d in devices.devices
        ]
        adhoc_session["has_next"] = False

        return adhoc_session

    async def _handle_add_device1(self, iq: Iq, adhoc_session: dict[str, Any]):
        user = user_store.get_by_stanza(iq)
        if user is None:
            raise XMPPError("subscription-required")

        form = self["xep_0004"].make_form(
            "form", "Link a new device to your signal account"
        )
        form.add_field(
            var="uri",
            ftype="text-single",
            label="Linking URI. Use a QR code reader app to get it from official signal clients.",
            required=True,
        )

        adhoc_session["payload"] = form
        adhoc_session["has_next"] = True
        adhoc_session["next"] = self._handle_add_device2

        return adhoc_session

    async def _handle_add_device2(self, stanza: Form, adhoc_session: dict[str, Any]):
        user = user_store.get_by_jid(adhoc_session["from"])
        if user is None:
            raise XMPPError("subscription-required")

        values = stanza.get_values()
        uri = values.get("uri")

        await (await self.signal).add_device(
            account=user.registration_form["phone"], uri=uri
        )

        adhoc_session["notes"] = [
            (
                "info",
                "Your new device is now correctly linked to your signal account",
            )
        ]
        adhoc_session["has_next"] = False

        return adhoc_session

    @staticmethod
    async def _chat_command_add_device(
        *args, msg: Message, session: Optional["Session"] = None
    ):
        if session is None:
            msg.reply("I don't know you, so don't talk to me").send()
            return
        if len(args) == 0:
            uri = await session.input("URI?")
        elif len(args) > 1:
            msg.reply("Syntax error! Use 'add_device [LINKING_URI]'").send()
            return
        else:
            uri = args[0]
        await session.add_device(uri)

    @staticmethod
    async def _chat_command_get_identities(
        *args, msg: Message, session: Optional["Session"] = None
    ):
        log.debug("ARGS: %s", args)
        if session is None:
            msg.reply("I don't know you, so don't talk to me").send()
            return
        if len(args) == 0:
            phone = await session.input("phone number?")
        elif len(args) > 1:
            msg.reply("Syntax error! Use 'get_identities [PHONE_NUMBER]'").send()
            return
        else:
            phone = args[0]
        await session.contacts.by_phone(phone).get_identities()

    async def validate(
        self, user_jid: JID, registration_form: dict[str, Optional[str]]
    ):
        phone = registration_form.get("phone")
        for u in user_store.get_all():
            if u.registration_form.get("phone") == phone:
                raise XMPPError(
                    "not-allowed",
                    text="Someone is already using this phone number on this server.\n",
                )
        if registration_form.get("device") == "primary" and not registration_form.get(
            "name"
        ):
            raise ValueError(txt.NAME_REQUIRED)

    async def unregister(self, user: GatewayUser):
        try:
            await (await self.signal).delete_account(
                account=user.registration_form.get("phone"), server=False
            )
        except sigexc.NoSuchAccountError:
            # if user unregisters before completing the registration process,
            # NoSuchAccountError is raised by signald
            pass

        log.info("Removed user: %s", user)


# noinspection PyPep8Naming,GrazieInspection
class Signal(SignaldAPI):
    """
    Extends :class:`.SignaldAPI` with handlers for events we are interested in.
    """

    def __init__(self, xmpp: Gateway):
        super().__init__()
        self.sessions_by_phone = xmpp.sessions_by_phone

    async def handle_ListenerState(self, state: sigapi.ListenerStatev1, payload):
        """
        Connection state for an account.

        :param state: State of the connection
        :param payload: The raw payload sent by signald
        """

        phone = payload["account"]
        if state.connected:
            session = self.sessions_by_phone[phone]
            await session.add_contacts_to_roster()

    async def handle_IncomingMessage(self, msg: sigapi.IncomingMessagev1, _payload):
        """
        Dispatch a signald message to the proper session.

        Can be a lot of other things than an actual message, still need to figure
        things out to cover all cases.

        :param msg: the data!
        :param _payload:
        """
        session = self.sessions_by_phone[msg.account]
        await session.on_signal_message(msg)


def get_parser():
    parser = ArgumentParser()
    parser.add_argument("--socket", default="/signald/signald.sock")
    return parser


log = logging.getLogger(__name__)



import asyncio
import json
import logging
import pprint
import ssl
import time
from dataclasses import dataclass
from enum import Enum

import aiohttp


class EventType(str, Enum):
    AddedToTeam = "added_to_team"
    AuthenticationChallenge = "authentication_challenge"
    ChannelConverted = "channel_converted"
    ChannelCreated = "channel_created"
    ChannelDeleted = "channel_deleted"
    ChannelMemberUpdated = "channel_member_updated"
    ChannelUpdated = "channel_updated"
    ChannelViewed = "channel_viewed"
    ConfigChanged = "config_changed"
    DeleteTeam = "delete_team"
    DirectAdded = "direct_added"
    EmojiAdded = "emoji_added"
    EphemeralMessage = "ephemeral_message"
    GroupAdded = "group_added"
    Hello = "hello"
    LeaveTeam = "leave_team"
    LicenseChanged = "license_changed"
    MemberroleUpdated = "memberrole_updated"
    NewUser = "new_user"
    PluginDisabled = "plugin_disabled"
    PluginEnabled = "plugin_enabled"
    PluginStatusesChanged = "plugin_statuses_changed"
    PostDeleted = "post_deleted"
    PostEdited = "post_edited"
    PostUnread = "post_unread"
    Posted = "posted"
    PreferenceChanged = "preference_changed"
    PreferencesChanged = "preferences_changed"
    PreferencesDeleted = "preferences_deleted"
    ReactionAdded = "reaction_added"
    ReactionRemoved = "reaction_removed"
    Response = "response"
    RoleUpdated = "role_updated"
    StatusChange = "status_change"
    Typing = "typing"
    UpdateTeam = "update_team"
    UserAdded = "user_added"
    UserRemoved = "user_removed"
    UserRoleUpdated = "user_role_updated"
    UserUpdated = "user_updated"
    DialogOpened = "dialog_opened"
    ThreadUpdated = "thread_updated"
    ThreadFollowChanged = "thread_follow_changed"
    ThreadReadChanged = "thread_read_changed"

    # not in the https://api.mattermost.com
    SidebarCategoryUpdated = "sidebar_category_updated"

    Unknown = "__unknown__"


@dataclass
class MattermostEvent:
    type: EventType
    data: dict
    broadcast: dict
    left: dict

    def __str__(self):
        return (
            f"<{self.type}:"
            f" \ndata: {pprint.pformat(self.data)}"
            f" \nbroadcast: {pprint.pformat(self.broadcast)}"
            f" \nleft: {pprint.pformat(self.left)}"
            f">"
        )


class Websocket:
    def __init__(self, url, token):
        self.token = token
        self.url = url

        self._alive = False
        self._last_msg = 0

        self.ssl_verify = True
        self.keep_alive = True
        self.keep_alive_delay = 30
        self.websocket: asyncio.Future[
            aiohttp.ClientWebSocketResponse
        ] = asyncio.get_event_loop().create_future()
        self._futures: dict[int, asyncio.Future[dict]] = {}
        self._seq_cursor = 0

    async def connect(self, event_handler):
        """
        Connect to the websocket and authenticate it.
        When the authentication has finished, start the loop listening for messages,
        sending a ping to the server to keep the connection alive.
        :param event_handler: Every websocket event will be passed there. Takes one argument.
        :type event_handler: Function(message)
        :return:
        """
        context = ssl.create_default_context(purpose=ssl.Purpose.CLIENT_AUTH)
        if not self.ssl_verify:
            context.verify_mode = ssl.CERT_NONE

        url = self.url
        self._alive = True

        while True:
            try:
                kw_args = {}
                async with aiohttp.ClientSession() as session:
                    async with session.ws_connect(
                        url,
                        ssl=context,
                        **kw_args,
                    ) as websocket:
                        self.websocket.set_result(websocket)
                        await self._authenticate_websocket(websocket)
                        while self._alive:
                            try:
                                await self._start_loop(websocket, event_handler)
                            except aiohttp.ClientError:
                                break
                        if (not self.keep_alive) or (not self._alive):
                            break
            except Exception as e:
                log.exception(
                    f"Failed to establish websocket connection: {type(e)} thrown"
                )
                await asyncio.sleep(self.keep_alive_delay)

    async def _start_loop(self, websocket, event_handler):
        """
        We will listen for websockets events, sending a heartbeats on a timer.
        If we don't the webserver would close the idle connection,
        forcing us to reconnect.
        """
        log.debug("Starting websocket loop")
        keep_alive = asyncio.create_task(self._do_heartbeats(websocket))
        log.debug("Waiting for messages on websocket")
        while self._alive:
            message = await websocket.receive_str()
            d = json.loads(message)
            self._last_msg = time.time()
            if (seq := d.get("seq_reply")) is None:
                await handle_event(d, event_handler)
            else:
                try:
                    self._futures.pop(seq).set_result(d)
                except KeyError:
                    log.warning("Ignoring %s", d)
        log.debug("cancelling heartbeat task")
        keep_alive.cancel()
        try:
            await keep_alive
        except asyncio.CancelledError:
            pass

    async def _do_heartbeats(self, websocket):
        """
        This is a little complicated, but we only need to pong the websocket if
        we haven't received a message inside the timeout window.
        Since messages can be received, while we are waiting we need to check
        after sleep.
        """
        timeout = 30
        while True:
            since_last_msg = time.time() - self._last_msg
            next_timeout = (
                timeout - since_last_msg if since_last_msg <= timeout else timeout
            )
            await asyncio.sleep(next_timeout)
            if time.time() - self._last_msg >= timeout:
                log.debug("sending heartbeat...")
                await websocket.pong()
                self._last_msg = time.time()

    def disconnect(self):
        """Sets `self._alive` to False so the loop in `self._start_loop` will finish."""
        log.info("Disconnecting websocket")
        self._alive = False

    async def _authenticate_websocket(self, websocket):
        """
        Sends an authentication challenge over a websocket.
        This is not needed when we just send the cookie we got on login
        when connecting to the websocket.
        """
        log.debug("Authenticating websocket")
        json_data = json.dumps(
            {
                "seq": 1,
                "action": "authentication_challenge",
                "data": {"token": self.token},
            }
        )
        await websocket.send_str(json_data)
        while True:
            message = await websocket.receive_str()
            status = json.loads(message)
            log.debug(status)
            if ("event" in status and status["event"] == "hello") and (
                "seq" in status and status["seq"] == 0
            ):
                log.info("Websocket authentication OK")
                return True
            log.error("Websocket authentication failed")

    async def user_typing(self, channel_id):
        seq = self._seq_cursor
        self._seq_cursor += 1
        f = self._futures[seq] = asyncio.get_event_loop().create_future()
        payload = json.dumps(
            {
                "seq": seq,
                "action": "user_typing",
                "data": {"channel_id": channel_id},
            }
        )
        log.debug("Sending %s", payload)
        await (await self.websocket).send_str(payload)
        r = await f
        log.debug("Confirmation %s", r)


async def handle_event(d, event_handler):
    if "event" in d:
        raw_data = d.pop("data")
        data = {}

        for k, v in raw_data.items():
            try:
                data[k] = json.loads(v)
            except (json.JSONDecodeError, TypeError):
                data[k] = v

        try:
            event = EventType(d.pop("event"))
        except ValueError:
            event = EventType.Unknown
        bro = d.pop("broadcast")
        await event_handler(MattermostEvent(event, data, bro, d))


log = logging.getLogger(__name__)



"""
A pseudo legacy network, to easily test things
"""

import asyncio
import logging
import uuid
from pathlib import Path
from typing import Any, Optional

from slixmpp import JID

from slidge import *

ASSETS_DIR = Path(__file__).parent.parent.parent / "assets"


class Gateway(BaseGateway):
    COMPONENT_NAME = "The great legacy network (slidge)"
    COMPONENT_AVATAR = ASSETS_DIR / "gateway.png"
    REGISTRATION_INSTRUCTIONS = (
        "Only username 'n' is accepted and only 'baba' and 'bibi' contacts exist.\n"
        "You can use any password you want."
    )
    REGISTRATION_FIELDS = list(BaseGateway.REGISTRATION_FIELDS) + [
        FormField(
            var="something_else",
            label="Some optional stuff not covered by jabber:iq:register",
            required=False,
            private=False,
        ),
        FormField(
            var="device",
            type="list-single",
            label="What do you want to do?",
            options=[
                {"label": "Choice #1", "value": "choice1"},
                {"label": "Choice #2", "value": "choice2"},
            ],
            required=True,
        ),
    ]

    async def validate(
        self, user_jid: JID, registration_form: dict[str, Optional[str]]
    ):
        if registration_form["username"] != "n":
            raise ValueError("Y a que N!")


class Session(BaseSession[LegacyContact, LegacyRoster, Gateway]):
    def __init__(self, user):
        super(Session, self).__init__(user)
        self.counter = 0
        self.xmpp.loop.create_task(self.backfill())

    async def backfill(self):
        self.log.debug("CARBON")
        i = uuid.uuid1()

        self.contacts.by_legacy_id("bibi").carbon(
            f"Sent by the component on behalf of the user, but this does not seem to reach MAM? {i}",
            legacy_id=i,
        )

    async def paused(self, c: LegacyContact):
        pass

    async def correct(self, text: str, legacy_msg_id: Any, c: LegacyContact):
        pass

    async def login(self):
        log.debug("Logging in user: %s", self.user)
        self.send_gateway_status("Connecting...", show="dnd")
        await asyncio.sleep(1)
        self.send_gateway_status("Connected")
        for b, a in zip(BUDDIES, AVATARS):
            c = self.contacts.by_legacy_id(b.lower())
            c.name = b.title()
            c.avatar = a
            await c.add_to_roster()
            c.online("I am not a real person, so what?")
        return "You can talk to your fake friends now"

    async def logout(self):
        log.debug("User has disconnected")

    async def send_text(self, t: str, c: LegacyContact, *, reply_to_msg_id=None):
        i = self.counter
        self.counter = i + 1
        self.xmpp.loop.create_task(self.later(c, i))

        if t == "crash":
            raise RuntimeError("PANIC!!!")
        return i

    async def send_file(self, u: str, c: LegacyContact, *, reply_to_msg_id=None) -> int:
        i = self.counter
        self.counter = i + 1
        c.send_text(u)
        await c.send_file(ASSETS_DIR / "buddy1.png")
        return i

    async def later(self, c: LegacyContact, trigger_msg_id: int):
        i = self.counter - 1
        await asyncio.sleep(1)
        c.received(i)
        await asyncio.sleep(1)
        c.active()
        await asyncio.sleep(1)
        c.displayed(i)
        await asyncio.sleep(1)
        c.ack(i)
        await asyncio.sleep(1)
        c.composing()
        await asyncio.sleep(1)
        c.paused()
        await asyncio.sleep(1)
        c.composing()
        await asyncio.sleep(1)
        c.send_text("OK", legacy_msg_id=i, reply_to_msg_id=trigger_msg_id)
        await asyncio.sleep(1)
        i = uuid.uuid1().int
        c.send_text("I will retract this", legacy_msg_id=i)
        c.retract(i)
        c.inactive()

    async def active(self, c: LegacyContact):
        log.debug("User is active for contact %s", c)

    async def inactive(self, c: LegacyContact):
        log.debug("User is inactive for contact %s", c)

    async def composing(self, c: LegacyContact):
        log.debug("User is composing for contact %s", c)

    async def displayed(self, legacy_msg_id: int, c: LegacyContact):
        log.debug("Message #%s was read by the user", legacy_msg_id)

    async def search(self, form_values: dict[str, str]):
        if form_values["first"] == "bubu":
            return SearchResult(
                fields=[
                    FormField("name", label="Name"),
                    FormField("jid", type="jid-single"),
                ],
                items=[{"name": "bubu", "jid": f"bubu@{self.xmpp.boundjid.bare}"}],
            )

    async def react(self, legacy_msg_id, emojis, c):
        c.react(legacy_msg_id, "♥")

    async def retract(self, legacy_msg_id, c):
        log.debug("User has retracted their msg: '%s' (sent to '%s')", legacy_msg_id, c)


BUDDIES = ["baba", "bibi"]
AVATARS = []

with (ASSETS_DIR / "buddy1.png").open("rb") as fp:
    AVATARS.append(fp.read())

with (ASSETS_DIR / "buddy2.png").open("rb") as fp:
    AVATARS.append(fp.read())

log = logging.getLogger(__name__)



"""
This module extends slixmpp.ComponentXMPP to make writing new LegacyClients easier
"""
import asyncio
import hashlib
import logging
import re
import tempfile
from asyncio import Future
from pathlib import Path
from typing import Any, Generic, Iterable, Optional, Type, TypeVar

import aiohttp
import qrcode
from slixmpp import JID, ComponentXMPP, Iq, Message
from slixmpp.exceptions import IqError, IqTimeout, XMPPError
from slixmpp.types import MessageTypes

from ..util import ABCSubclassableOnceAtMost, FormField
from ..util.db import GatewayUser, RosterBackend, user_store
from ..util.types import AvatarType
from .session import BaseSession, SessionType


class BaseGateway(
    Generic[SessionType], ComponentXMPP, metaclass=ABCSubclassableOnceAtMost
):
    """
    Must be subclassed by a plugin to set up various aspects of the XMPP
    component behaviour, such as its display name or its registration process.

    On slidge launch, a singleton is instantiated, and it will be made available
    to public classes such :class:`.LegacyContact` or :class:`.BaseSession` as the
    ``.xmpp`` attribute.
    Since it inherits from :class:`slixmpp.componentxmpp.ComponentXMPP`, this gives you a hand
    on low-level XMPP interactions via slixmpp plugins, e.g.:

    .. code-block:: python

        self.send_presence(
            pfrom="somebody@component.example.com",
            pto="someonwelse@anotherexample.com",
        )

    However, you should not need to do so often since the classes of the plugin
    API provides higher level abstractions around most commonly needed use-cases, such
    as sending messages, or displaying a custom status.

    """

    REGISTRATION_FIELDS: Iterable[FormField] = [
        FormField(var="username", label="User name", required=True),
        FormField(var="password", label="Password", required=True, private=True),
    ]
    """
    Iterable of fields presented to the gateway user when registering using :xep:`0077`
    `extended <https://xmpp.org/extensions/xep-0077.html#extensibility>`_ by :xep:`0004`.
    """
    REGISTRATION_INSTRUCTIONS: str = "Enter your credentials"
    """
    The text presented to a user that wants to register (or modify) their legacy account
    configuration.
    """

    COMPONENT_NAME: str = NotImplemented
    """Name of the component, as seen in service discovery by XMPP clients"""
    COMPONENT_TYPE: Optional[str] = ""
    """Type of the gateway, should ideally follow https://xmpp.org/registrar/disco-categories.html"""
    COMPONENT_AVATAR: Optional[AvatarType] = None
    """
    Path, bytes or URL used by the component as an avatar.
    """

    ROSTER_GROUP: str = "slidge"
    """
    Roster entries added by the plugin in the user's roster will be part of the group specified here.
    """

    SEARCH_FIELDS: Iterable[FormField] = [
        FormField(var="first", label="First name", required=True),
        FormField(var="last", label="Last name", required=True),
    ]
    """
    Fields used for searching items via the component, through :xep:`0055` (jabber search).
    A common use case is to allow users to search for legacy contacts by something else than
    their usernames, eg their phone number.
    
    Plugins should implement search by overriding :meth:`.BaseSession.search`, effectively
    restricting search to registered users by default.
    """
    SEARCH_TITLE: str = "Search for legacy contacts"
    """
    Title of the search form.
    """
    SEARCH_INSTRUCTIONS: str = ""
    """
    Instructions of the search form.
    """

    _BASE_CHAT_COMMANDS = {
        "find": "_chat_command_search",
        "help": "_chat_command_help",
        "register": "_chat_command_register",
    }
    CHAT_COMMANDS: dict[str, str] = {}
    """
    Keys of this dict can be used to trigger a command by a simple chat message to the gateway
    component. Extra words after the key are passed as *args to the handler. Values of the dict
    are strings, and handlers are resolved using ``getattr()`` on the :class:`.BaseGateway`
    instance.
    
    Handlers are coroutines with following signature:
    
    .. code-block::python
    
        async def _chat_command_xxx(*args, msg: Message, session: Optional[Session] = None)
            ...
    
    The original :class:`slixmpp.stanza.Message` is also passed to the handler as the
    msg kwarg. If the command comes from a registered gateway user, its session attribute is also
    passed to the handler.
    """

    def __init__(self, args):
        """

        :param args: CLI arguments parsed by :func:`.slidge.__main__.get_parser`
        """
        super().__init__(
            args.jid,
            args.secret,
            args.server,
            args.port,
            plugin_whitelist=SLIXMPP_PLUGINS,
            plugin_config={
                "xep_0077": {
                    "form_fields": None,
                    "form_instructions": self.REGISTRATION_INSTRUCTIONS,
                },
                "xep_0100": {
                    "component_name": self.COMPONENT_NAME,
                    "user_store": user_store,
                    "type": self.COMPONENT_TYPE,
                },
                "xep_0184": {
                    "auto_ack": False,
                    "auto_request": True,
                },
                "xep_0363": {
                    "upload_service": args.upload_service,
                },
            },
        )
        self.loop.set_exception_handler(self.__exception_handler)

        self.home_dir = Path(args.home_dir)
        self._jid_validator = re.compile(args.user_jid_validator)
        self._config = args
        self.no_roster_push = args.no_roster_push

        self._session_cls: Type[SessionType] = BaseSession.get_unique_subclass()
        self._session_cls.xmpp = self

        self._get_session_from_stanza = self._session_cls.from_stanza
        self._get_session_from_user = self._session_cls.from_user
        self.register_plugins()
        self.__register_slixmpp_api()
        self.__register_handlers()
        self._input_futures: dict[str, Future] = {}

        self._chat_commands = {
            k: getattr(self, v)
            for k, v in (self._BASE_CHAT_COMMANDS | self.CHAT_COMMANDS).items()
        }

    @staticmethod
    def __exception_handler(loop: asyncio.AbstractEventLoop, context):
        loop.default_exception_handler(context)  # prints a standard python traceback
        exc = context.get("exception")
        if exc is None:
            log.warning("No exception in this context: %s", context)
        else:
            log.exception(exc)
        loop.stop()

    def exception(self, exception: Exception):
        """
        Stop the event loop and exit on unhandled exception.

        The default :class:slixmpp.basexmpp.BaseXMPP` behaviour is just to
        log the exception, but we want to avoid undefined behaviour.

        :param exception: An unhandled :class:`Exception` object.
        """
        if isinstance(exception, IqError):
            iq = exception.iq
            log.error("%s: %s", iq["error"]["condition"], iq["error"]["text"])
            log.warning("You should catch IqError exceptions")
        elif isinstance(exception, IqTimeout):
            iq = exception.iq
            log.error("Request timed out: %s", iq)
            log.warning("You should catch IqTimeout exceptions")
        elif isinstance(exception, SyntaxError):
            # Hide stream parsing errors that occur when the
            # stream is disconnected (they've been handled, we
            # don't need to make a mess in the logs).
            pass
        else:
            log.exception(exception)
            self.loop.stop()
            exit(1)

    def __register_slixmpp_api(self):
        self["xep_0077"].api.register(
            user_store.get,
            "user_get",
        )
        self["xep_0077"].api.register(
            user_store.remove,
            "user_remove",
        )
        self["xep_0077"].api.register(
            self._make_registration_form,
            "make_registration_form",
        )
        self["xep_0077"].api.register(self._user_validate, "user_validate")
        self["xep_0077"].api.register(self._user_modify, "user_modify")

        self["xep_0055"].api.register(self._search_get_form, "search_get_form")
        self["xep_0055"].api.register(self._search_query, "search_query")

        self.roster.set_backend(RosterBackend)

    def __register_handlers(self):
        self.add_event_handler("session_start", self.__on_session_start)
        self.add_event_handler("disconnected", self.connect)
        self.add_event_handler("gateway_message", self._on_gateway_message_private)
        self.add_event_handler("user_register", self._on_user_register)
        self.add_event_handler("user_unregister", self._on_user_unregister)
        get_session = self._get_session_from_stanza

        # fmt: off
        async def msg(m): await get_session(m).send_from_msg(m)
        async def disp(m): await get_session(m).displayed_from_msg(m)
        async def active(m): await get_session(m).active_from_msg(m)
        async def inactive(m): await get_session(m).inactive_from_msg(m)
        async def composing(m): await get_session(m).composing_from_msg(m)
        async def paused(m): await get_session(m).paused_from_msg(m)
        async def correct(m): await get_session(m).correct_from_msg(m)
        async def react(m): await get_session(m).react_from_msg(m)
        async def retract(m): await get_session(m).retract_from_msg(m)
        # fmt: on

        self.add_event_handler("legacy_message", msg)
        self.add_event_handler("marker_displayed", disp)
        self.add_event_handler("chatstate_active", active)
        self.add_event_handler("chatstate_inactive", inactive)
        self.add_event_handler("chatstate_composing", composing)
        self.add_event_handler("chatstate_paused", paused)
        self.add_event_handler("message_correction", correct)
        self.add_event_handler("reactions", react)
        self.add_event_handler("message_retract", retract)

    async def __on_session_start(self, event):
        log.debug("Gateway session start: %s", event)

        # prevents XMPP clients from considering the gateway as an HTTP upload
        await self["xep_0030"].del_feature(
            feature="urn:xmpp:http:upload:0", jid=self.boundjid.bare
        )

        self.__add_adhoc_commands()
        self.add_adhoc_commands()
        await self.set_vcard_avatar(self.boundjid.bare, self.COMPONENT_AVATAR)

        for user in user_store.get_all():
            # TODO: before this, we should check if the user has removed us from their roster
            #       while we were offline and trigger unregister from there. Presence probe does not seem
            #       to work in this case, there must be another way. privileged entity could be used
            #       as last resort.
            self.send_presence(
                pto=user.bare_jid, ptype="probe"
            )  # ensure we get all resources for user
            session = self._session_cls.from_user(user)
            self.loop.create_task(self._login_wrap(session))

        log.info("Slidge has successfully started")

    @staticmethod
    async def _login_wrap(session: "SessionType"):
        session.send_gateway_status("Logging in…", show="dnd")
        try:
            status = await session.login()
        except Exception as e:
            session.send_gateway_status(f"Could not login: {e}", show="busy")
            session.send_gateway_message(
                f"You are not connected to this gateway! "
                f"Maybe this message will tell you why: {e}"
            )
        else:
            if status is None:
                session.send_gateway_status("Logged in", show="chat")
            else:
                session.send_gateway_status(status, show="chat")

    def __add_adhoc_commands(self):
        # TODO: this should only be advertised to admins
        # Not a big deal since we need to check if 'from' is an admin in the handler
        # anyway, BUT it would be nice if this simply does not show up in the list
        # of available commands for regular users.
        self["xep_0050"].add_command(
            node="info", name="List registered users", handler=self._handle_info
        )

    def _handle_info(self, iq: Iq, session: dict[str, Any]):
        """
        List registered users for admins
        """
        if iq.get_from().bare not in self._config.admins:
            raise XMPPError("not-authorized")
        form = self["xep_0004"].make_form("result", "Component info")
        form.add_field(
            ftype="jid-multi",
            label="Users",
            value=[u.bare_jid for u in user_store.get_all()],
        )

        session["payload"] = form
        session["has_next"] = False

        return session

    async def _make_registration_form(self, _jid, _node, _ifrom, iq: Iq):
        if not self._jid_validator.match(iq.get_from().bare):
            raise XMPPError(condition="not-allowed")

        reg = iq["register"]
        user = user_store.get_by_stanza(iq)
        log.debug("User found: %s", user)

        form = reg["form"]
        form.add_field(
            "FORM_TYPE",
            ftype="hidden",
            value="jabber:iq:register",
        )
        form["title"] = f"Registration to '{self.COMPONENT_NAME}'"
        form["instructions"] = self.REGISTRATION_INSTRUCTIONS

        if user is not None:
            reg["registered"] = False
            form.add_field(
                "remove",
                label="Remove my registration",
                required=True,
                ftype="boolean",
                value=False,
            )

        for field in self.REGISTRATION_FIELDS:
            if field.var in reg.interfaces:
                val = None if user is None else user.get(field.var)
                if val is None:
                    reg.add_field(field.var)
                else:
                    reg[field.var] = val

        reg["instructions"] = self.REGISTRATION_INSTRUCTIONS

        for field in self.REGISTRATION_FIELDS:
            form.add_field(
                field.var,
                label=field.label,
                required=field.required,
                ftype=field.type,
                options=field.options,
                value=field.value if user is None else user.get(field.var, field.value),
            )

        reply = iq.reply()
        reply.set_payload(reg)
        return reply

    async def _user_prevalidate(self, ifrom: JID, form_dict: dict[str, Optional[str]]):
        """
        Pre validate a registration form using the content of self.REGISTRATION_FIELDS
        before passing it to the plugin custom validation logic.
        """
        for field in self.REGISTRATION_FIELDS:
            if field.required and not form_dict.get(field.var):
                raise ValueError(f"Missing field: '{field.label}'")

        await self.validate(ifrom, form_dict)

    async def _user_validate(
        self, _gateway_jid, _node, ifrom: JID, form_dict: dict[str, Optional[str]]
    ):
        """
        SliXMPP internal API stuff
        """
        log.debug("User validate: %s", ifrom.bare)
        if not self._jid_validator.match(ifrom.bare):
            raise XMPPError(condition="not-allowed")
        await self._user_prevalidate(ifrom, form_dict)
        log.info("New user: %s", ifrom.bare)
        user_store.add(ifrom, form_dict)

    async def _user_modify(
        self, _gateway_jid, _node, ifrom: JID, form_dict: dict[str, Optional[str]]
    ):
        """
        SliXMPP internal API stuff
        """
        user = user_store.get_by_jid(ifrom)
        log.debug("Modify user: %s", user)
        await self._user_prevalidate(ifrom, form_dict)
        user_store.add(ifrom, form_dict)

    async def _on_user_register(self, iq: Iq):
        session = self._get_session_from_stanza(iq)
        for jid in self._config.admins:
            self.send_message(
                mto=jid,
                mbody=f"{iq.get_from()} has registered",
                mtype="headline",
                mfrom=self.boundjid.bare,
            )
        await session.login()

    async def _on_user_unregister(self, iq: Iq):
        # Mypy: "Type[SessionType?]" has no attribute "kill_by_jid"
        # I don't understand why ^ this question mark...
        kill = self._session_cls.kill_by_jid  # type: ignore
        await kill(iq.get_from())

    async def _search_get_form(self, _gateway_jid, _node, ifrom: JID, iq: Iq):
        """
        Prepare the search form using self.SEARCH_FIELDS
        """
        user = user_store.get_by_jid(ifrom)
        if user is None:
            raise XMPPError(text="Search is only allowed for registered users")

        reply = iq.reply()
        form = reply["search"]["form"]
        form["title"] = self.SEARCH_TITLE
        form["instructions"] = self.SEARCH_INSTRUCTIONS
        for field in self.SEARCH_FIELDS:
            form.add_field(**field.dict())
        return reply

    async def _search_query(self, _gateway_jid, _node, ifrom: JID, iq: Iq):
        """
        Handles a search request
        """
        user = user_store.get_by_jid(ifrom)
        if user is None:
            raise XMPPError(text="Search is only allowed for registered users")

        result = await self._get_session_from_stanza(iq).search(
            iq["search"]["form"].get_values()
        )

        if not result:
            raise XMPPError("item-not-found", text="Nothing was found")

        reply = iq.reply()
        form = reply["search"]["form"]
        for field in result.fields:
            form.add_reported(field.var, label=field.label, type=field.type)
        for item in result.items:
            form.add_item(item)
        return reply

    async def _chat_command_search(
        self, *args, msg: Message, session: Optional[SessionType] = None
    ):
        if session is None:
            msg.reply("Register to the gateway first!")
            return

        search_form = {}
        if len(args) == 0:
            for field in self.SEARCH_FIELDS:
                search_form[field.var] = await session.input(
                    (field.label or field.var) + "?"
                )
        else:
            for field, arg in zip(self.SEARCH_FIELDS, args):
                search_form[field.var] = arg

        results = await session.search(search_form)
        if results is None:
            session.send_gateway_message("No results!")
            return

        result_fields = results.fields
        for result in results.items:
            text = ""
            for f in result_fields:
                if f.type == "jid-single":
                    text += f"xmpp:{result[f.var]}\n"
                else:
                    text += f"{f.label}: {result[f.var]}\n"
            session.send_gateway_message(text)

    async def _chat_command_help(
        self, *_args, msg: Message, session: Optional[SessionType]
    ):
        if session is None:
            msg.reply("Register to the gateway first!").send()
        else:
            t = "|".join(self._chat_commands.keys())
            log.debug("In help: %s", t)
            msg.reply(f"Available commands: {t}").send()

    async def _chat_command_register(
        self, *args, msg: Message, session: Optional[SessionType]
    ):
        if session is not None:
            msg.reply("You are already registered to this gateway").send()
            return

        jid = msg.get_from()

        if not self._jid_validator.match(jid.bare):
            msg.reply("You are not allowed to register to this gateway").send()
            return

        form: dict[str, Optional[str]] = {}
        for field in self.REGISTRATION_FIELDS:
            text = field.label or field.var
            if field.value != "":
                text += f" (default: '{field.value}')"
            if not field.required:
                text += " (optional, reply with '.' to skip)"
            if (options := field.options) is not None:
                for option in options:
                    label = option["label"]
                    value = option["value"]
                    text += f"\n{label}: reply with '{value}'"

            while True:
                ans = await self.input(jid, text + "?")
                if ans == "." and not field.required:
                    form[field.var] = None
                    break
                else:
                    if (options := field.options) is not None:
                        valid_choices = [x["value"] for x in options]
                        if ans not in valid_choices:
                            continue
                    form[field.var] = ans
                    break

        try:
            await self.validate(jid, form)
            await self["xep_0077"].api["user_validate"](None, None, jid, form)
        except (ValueError, XMPPError) as e:
            msg.reply(f"Something went wrong: {e}").send()
        else:
            self.event("user_register", msg)
            msg.reply(f"Success!").send()

    def add_adhoc_commands(self):
        """
        Override this if you want to provide custom adhoc commands (:xep:`0050`)
        for your plugin, using :class:`slixmpp.plugins.xep_0050.XEP_0050`

        Basic example:

        .. code-block:python

            def add_adhoc_commands(self):
                self["xep_0050"].add_command(
                    node="account_info",
                    name="Account Information",
                    handler=self.handle_account_info
                )

            async def handle_account_info(self, iq: Iq, adhoc_session: dict[str, Any]):
                # beware, 'adhoc_session' is not a slidge session!
                user = user_store.get_by_stanza(iq)

                if user is None:
                    raise XMPPError("subscription-required")

                form = self["xep_0004"].make_form("result", "Account info")
                form.add_field(
                    label="Credits",
                    value=await FakeLegacyClient.get_credits(user.registration_form['username']),
                )

                adhoc_session["payload"] = form
                adhoc_session["has_next"] = False

                return session
        """
        pass

    def config(self, argv: list[str]):
        """
        Override this to access CLI args to configure the slidge plugin

        :param argv: CLI args that were not parsed by the slidge main entrypoint parser
        :func:`slidge.__main__.get_parser`
        """
        pass

    async def validate(
        self, user_jid: JID, registration_form: dict[str, Optional[str]]
    ):
        """
        Validate a registration form from a user.

        Since :xep:`0077` is pretty limited in terms of validation, it is OK to validate
        anything that looks good here and continue the legacy auth process via direct messages
        to the user (using :meth:`.BaseGateway.input` for instance)

        :param user_jid: JID of the user that has just registered
        :param registration_form: A dict where keys are the :attr:`.FormField.var` attributes
         of the :attr:`.BaseGateway.REGISTRATION_FIELDS` iterable
        """
        pass

    async def unregister(self, user: GatewayUser):
        """
        Optionally override this if you need to clean additional
        stuff after a user has been removed from the permanent user_store.

        :param user:
        :return:
        """
        pass

    async def set_vcard_avatar(self, jid: JID, avatar: AvatarType):
        """
        Set an avatar for a specific JID.

        For contacts, :attr:`.LegacyContact.avatar` property

        :param jid: JID to associate the avatar with
        :param avatar: bytes, URL, or Path
        """
        vcard = self["xep_0054"].make_vcard()
        if avatar is not None:
            if isinstance(avatar, bytes):
                avatar_bytes = avatar
            elif isinstance(avatar, Path):
                with avatar.open("rb") as f:
                    avatar_bytes = f.read()
            elif isinstance(avatar, str):
                async with aiohttp.ClientSession() as session:
                    async with session.get(avatar) as response:
                        avatar_bytes = await response.read()
            else:
                raise TypeError("Avatar must be bytes, a Path or a str (URL)", avatar)
            vcard["PHOTO"]["BINVAL"] = avatar_bytes
            await self["xep_0153"].api["set_hash"](
                jid=jid, args=hashlib.sha1(avatar_bytes).hexdigest()
            )
        await self["xep_0054"].api["set_vcard"](
            jid=jid,
            args=vcard,
        )

    async def _on_gateway_message_private(self, msg: Message):
        """
        Called when an XMPP user (not necessarily registered as a gateway user) sends a direct message to
        the gateway.

        If you override this and still want :meth:`.BaseGateway.input` to work, make sure to include the try/except part.

        :param msg: Message sent by the XMPP user
        """
        try:
            f = self._input_futures.pop(msg.get_from().bare)
        except KeyError:
            text = msg["body"]
            command, *rest = text.split(" ")

            user = user_store.get_by_stanza(msg)
            if user is None:
                session = None
            else:
                session = self._get_session_from_user(user)

            handler = self._chat_commands.get(command)
            if handler is None:
                await self.on_gateway_message(msg, session=session)
            else:
                log.debug("Chat command handler: %s", handler)
                await handler(*rest, msg=msg, session=session)
        else:
            f.set_result(msg["body"])

    @staticmethod
    async def on_gateway_message(msg: Message, session: Optional[SessionType] = None):
        """
        Called when the gateway component receives a direct gateway message.

        Can be used to implement bot like commands, especially in conjunction with
        :meth:`.BaseGateway.input`

        :param msg:
        :param session: If the message comes from a registered gateway user, their :.BaseSession:
        """
        if session is None:
            r = msg.reply(
                body="I got that, but I'm not doing anything with it. I don't even know you!"
            )
        else:
            r = msg.reply(body="What? Type 'help' for the list of available commands.")
        r["type"] = "chat"
        r.send()

    async def input(
        self, jid: JID, text=None, mtype: MessageTypes = "chat", **msg_kwargs
    ) -> str:
        """
        Request arbitrary user input using a simple chat message, and await the result.

        You shouldn't need to call directly bust instead user :meth:`.BaseSession.input`
        to directly target a user.

        NB: When using this, the next message that the user sent to the component will
        not be transmitted to :meth:`.BaseGateway.on_gateway_message`, but rather intercepted.
        Await the coroutine to get its content.

        :param jid: The JID we want input from
        :param text: A prompt to display for the user
        :param mtype: Message type
        :return: The user's reply
        """
        if text is not None:
            self.send_message(
                mto=jid, mbody=text, mtype=mtype, mfrom=self.boundjid.bare, **msg_kwargs
            )
        f = self.loop.create_future()
        self._input_futures[jid.bare] = f
        await f
        return f.result()

    async def send_file(self, filename: str, **msg_kwargs):
        """
        Upload a file using :xep:`0363` and send the link as out of band (:xep:`0066`)
        content in a message.

        :param filename:
        :param msg_kwargs:
        :return:
        """
        url = await self["xep_0363"].upload_file(filename=filename)
        msg = self.make_message(**msg_kwargs)
        msg["oob"]["url"] = url
        msg["body"] = url
        msg.send()

    async def send_qr(self, text: str, **msg_kwargs):
        """
        Sends a QR Code to a JID

        :param text: The text that will be converted to a QR Code
        :param msg_kwargs: Optional additional arguments to pass to :meth:`.BaseGateway.send_file`,
            such as the recipient of the QR code.
        """
        qr = qrcode.make(text)
        with tempfile.NamedTemporaryFile(suffix=".png") as f:
            qr.save(f.name)
            await self.send_file(f.name, **msg_kwargs)

    def shutdown(self):
        """
        Called by the slidge entrypoint on normal exit.

        Sends offline presences from all contacts of all user sessions and from
        the gateway component itself.
        No need to call this manually, :func:`slidge.__main__.main` should take care of it.
        """
        log.debug("Shutting down")
        for user in user_store.get_all():
            session = self._session_cls.from_jid(user.jid)
            for c in session.contacts:
                c.offline()
            self.loop.create_task(session.logout())
            self.send_presence(ptype="unavailable", pto=user.jid)


GatewayType = TypeVar("GatewayType", bound=BaseGateway)


SLIXMPP_PLUGINS = [
    "xep_0050",  # Adhoc commands
    "xep_0054",  # vCard-temp
    "xep_0055",  # Jabber search
    "xep_0066",  # Out of Band Data
    "xep_0077",  # In-band registration
    "xep_0085",  # Chat state notifications
    "xep_0100",  # Gateway interaction
    "xep_0115",  # Entity capabilities
    "xep_0153",  # vCard-Based Avatars
    "xep_0184",  # Message Delivery Receipts
    "xep_0280",  # Carbons
    "xep_0308",  # Last message correction
    "xep_0333",  # Chat markers
    "xep_0334",  # Message Processing Hints
    "xep_0356",  # Privileged Entity
    "xep_0356_old",  # Privileged Entity (old namespace)
    "xep_0363",  # HTTP file upload
    "xep_0424",  # Message retraction
    "xep_0444",  # Message reactions
    "xep_0461",  # Message replies
]
log = logging.getLogger(__name__)



import logging
import typing
from argparse import Namespace
from datetime import datetime

import aiotdlib.api as tgapi
from slixmpp import JID, Iq
from slixmpp.exceptions import XMPPError

from slidge import *

from .config import get_parser

if typing.TYPE_CHECKING:
    from .session import Session

REGISTRATION_INSTRUCTIONS = """You can visit https://my.telegram.org/apps to get an API ID and an API HASH

This is the only tested login method, but other methods (password, bot token, 2FA...)
should work too, in theory at least.
"""


class Gateway(BaseGateway["Session"]):
    REGISTRATION_INSTRUCTIONS = REGISTRATION_INSTRUCTIONS
    REGISTRATION_FIELDS = [
        FormField(var="phone", label="Phone number", required=True),
        FormField(var="api_id", label="API ID", required=False),
        FormField(var="api_hash", label="API hash", required=False),
        FormField(var="", value="The fields below have not been tested", type="fixed"),
        FormField(var="bot_token", label="Bot token", required=False),
        FormField(var="first", label="First name", required=False),
        FormField(var="last", label="Last name", required=False),
    ]
    ROSTER_GROUP = "Telegram"
    COMPONENT_NAME = "Telegram (slidge)"
    COMPONENT_TYPE = "telegram"
    COMPONENT_AVATAR = "https://web.telegram.org/img/logo_share.png"

    SEARCH_FIELDS = [
        FormField(var="phone", label="Phone number", required=True),
    ]

    args: Namespace

    def config(self, argv: list[str]):
        Gateway.args = args = get_parser().parse_args(argv)
        if args.tdlib_path is None:
            args.tdlib_path = self.home_dir / "tdlib"

    async def validate(
        self, user_jid: JID, registration_form: dict[str, typing.Optional[str]]
    ):
        pass

    async def unregister(self, user):
        pass

    def add_adhoc_commands(self):
        self["xep_0050"].add_command(
            node="get_sessions",
            name="List active sessions",
            handler=self.adhoc_active_sessions1,
        )

    async def adhoc_active_sessions1(
        self, iq: Iq, adhoc_session: dict[str, typing.Any]
    ):
        user = user_store.get_by_stanza(iq)
        if user is None:
            raise XMPPError("subscription-required")
        session = self._session_cls.from_stanza(iq)

        form = self["xep_0004"].make_form("form", "Active telegram sessions")
        tg_sessions = (await session.list_sessions()).sessions
        form.add_field(
            "tg_session_id",
            ftype="list-single",
            label="Sessions",
            options=[{"label": f"{s.country}", "value": s.id} for s in tg_sessions],
        )

        adhoc_session["payload"] = form
        adhoc_session["next"] = self.adhoc_active_sessions2
        adhoc_session["has_next"] = True
        adhoc_session["tg_sessions"] = {s.id: s for s in tg_sessions}
        adhoc_session["slidge_session"] = session

        return adhoc_session

    async def adhoc_active_sessions2(self, form, adhoc_session: dict[str, typing.Any]):
        tg_session_id = int(form.get_values()["tg_session_id"])

        form = self["xep_0004"].make_form("form", "Telegram session info")
        tg_session: tgapi.Session = adhoc_session["tg_sessions"][str(tg_session_id)]
        for x in fmt_tg_session(tg_session):
            form.add_field(
                ftype="fixed",
                value=x,
            )
        if tg_session.is_current:
            adhoc_session["has_next"] = False
        else:
            form.add_field(
                "terminate", ftype="boolean", label="Terminate session", value="0"
            )
            form.add_field("tg_session_id", ftype="hidden", value=tg_session.id)
            adhoc_session["has_next"] = True
            adhoc_session["next"] = self.adhoc_active_sessions3

        adhoc_session["payload"] = form

        return adhoc_session

    async def adhoc_active_sessions3(self, form, adhoc_session: dict[str, typing.Any]):
        form_values = form.get_values()
        terminate = bool(int(form_values["terminate"]))

        if terminate:
            session: Session = adhoc_session["slidge_session"]
            await session.terminate_session(int(form_values["tg_session_id"]))
            info = "Session terminated."
        else:
            info = "Session not terminated."

        adhoc_session["notes"] = [("info", info)]
        adhoc_session["has_next"] = False

        return adhoc_session


def fmt_tg_session(s: tgapi.Session):
    return [
        f"Country: {s.country}",
        f"Region: {s.region}",
        f"Ip: {s.ip}",
        f"App: {s.application_name}",
        f"Device: {s.device_model}",
        f"Platform: {s.platform}",
        f"Since: {datetime.fromtimestamp(s.log_in_date).isoformat()}",
        f"Last seen: {datetime.fromtimestamp(s.last_active_date).isoformat()}",
    ]


log = logging.getLogger(__name__)



import slixmpp.plugins

from slidge.core.contact import LegacyContact, LegacyRoster
from slidge.core.gateway import BaseGateway
from slidge.core.session import BaseSession
from slidge.util.db import GatewayUser, user_store

from .util import (
    FormField,
    SearchResult,
    xep_0030,
    xep_0055,
    xep_0077,
    xep_0100,
    xep_0115,
    xep_0333,
    xep_0356,
    xep_0356_old,
    xep_0363,
    xep_0461,
)

slixmpp.plugins.__all__.extend(["xep_0055", "xep_0356", "xep_0356_old", "xep_0461"])

__all__ = [
    "BaseGateway",
    "BaseSession",
    "GatewayUser",
    "LegacyContact",
    "LegacyRoster",
    "FormField",
    "SearchResult",
    "user_store",
]



import functools
import logging
from typing import TYPE_CHECKING, Any, Generic, Literal, Optional, Type, TypeVar

from slixmpp import JID, Message
from slixmpp.exceptions import XMPPError

from ..core.contact import LegacyContactType, LegacyRoster, LegacyRosterType
from ..util import ABCSubclassableOnceAtMost, BiDict
from ..util.db import GatewayUser, user_store
from ..util.types import LegacyMessageType

if TYPE_CHECKING:
    from slidge import SearchResult
    from slidge.core.gateway import GatewayType
else:
    GatewayType = TypeVar("GatewayType")


def ignore_message_to_component(func):
    @functools.wraps(func)
    async def wrapped(self: T, msg: Message):
        if msg.get_to() != self.xmpp.boundjid.bare:
            return await func(self, msg)
        log.debug("Ignoring message to component: %s %s", self, msg)

    return wrapped


T = TypeVar("T", bound="BaseSession")


class BaseSession(
    Generic[LegacyContactType, LegacyRosterType, GatewayType],
    metaclass=ABCSubclassableOnceAtMost,
):
    """
    Represents a gateway user logged in to the network and performing actions.

    Will be instantiated automatically when a user sends an online presence to the gateway
    component, as per :xep:`0100`.

    Must be subclassed for a functional slidge plugin.
    """

    sent: BiDict[LegacyMessageType, str]
    """
    Since we cannot set the XMPP ID of messages sent by XMPP clients, we need to keep a mapping
    between XMPP IDs and legacy message IDs if we want to further refer to a message that was sent
    by the user. This also applies to 'carboned' messages, ie, messages sent by the user from
    the official client of a legacy network.
    """

    xmpp: "GatewayType"
    """
    The gateway instance singleton. Use it for low-level XMPP calls or custom methods that are not
    session-specific.
    """

    def __init__(self, user: GatewayUser):
        self._roster_cls: Type[
            LegacyRosterType
        ] = LegacyRoster.get_self_or_unique_subclass()

        self.log = logging.getLogger(user.bare_jid)

        self.user = user
        self.sent: BiDict[
            LegacyMessageType, str
        ] = BiDict()  # TODO: set a max size for this

        self.ignore_messages = set[str]()

        self.contacts: LegacyRosterType = self._roster_cls(self)
        self.post_init()

    @staticmethod
    def legacy_msg_id_to_xmpp_msg_id(legacy_msg_id: Any) -> str:
        """
        Convert a legacy msg ID to a valid XMPP msg ID.
        Needed for read marks and message corrections.

        The default implementation just converts the legacy ID to a :class:`str`,
        but this should be overridden in case some characters needs to be escaped,
        or to add some additional, legacy network-specific logic.

        :param legacy_msg_id:
        :return: Should return a string that is usable as an XMPP stanza ID
        """
        return str(legacy_msg_id)

    @staticmethod
    def xmpp_msg_id_to_legacy_msg_id(i: str) -> LegacyMessageType:
        """
        Convert a legacy XMPP ID to a valid XMPP msg ID.
        Needed for read marks and message corrections.

        The default implementation just converts the legacy ID to a :class:`str`,
        but this should be overridden in case some characters needs to be escaped,
        or to add some additional, legacy network-specific logic.

        The default implementation is an identity function

        :param i: The XMPP stanza ID
        :return: An ID that can be used to identify a message on the legacy network
        """
        return i

    @classmethod
    def _from_user_or_none(cls, user):
        if user is None:
            raise XMPPError(
                text="User not found", condition="subscription-required", etype="auth"
            )

        session = _sessions.get(user)
        if session is None:
            _sessions[user] = session = cls(user)
        return session

    @classmethod
    def from_user(cls, user):
        return cls._from_user_or_none(user)

    @classmethod
    def from_stanza(cls: Type[T], s) -> T:
        """
        Get a user's :class:`.LegacySession` using the "from" field of a stanza

        Meant to be called from :class:`BaseGateway` only.

        :param s:
        :return:
        """
        return cls._from_user_or_none(user_store.get_by_stanza(s))

    @classmethod
    def from_jid(cls: Type[T], jid: JID) -> T:
        """
        Get a user's :class:`.LegacySession` using its jid

        Meant to be called from :class:`BaseGateway` only.

        :param jid:
        :return:
        """
        return cls._from_user_or_none(user_store.get_by_jid(jid))

    @classmethod
    async def kill_by_jid(cls, jid: JID):
        """
        Terminate a user session.

        Meant to be called from :class:`BaseGateway` only.

        :param jid:
        :return:
        """
        log.debug("Killing session of %s", jid)
        for user, session in _sessions.items():
            if user.jid == jid.bare:
                break
        else:
            log.debug("Did not find a session for %s", jid)
            return
        for c in session.contacts:
            c.unsubscribe()
        await session.logout()
        await cls.xmpp.unregister(user)
        del _sessions[user]
        del user
        del session

    @ignore_message_to_component
    async def send_from_msg(self, m: Message):
        """
        Meant to be called from :class:`BaseGateway` only.

        :param m:
        :return:
        """
        if m["replace"][
            "id"
        ]:  # ignore last message correction (handled by a specific method)
            return
        if (i := m.get_id()) in self.ignore_messages:
            self.log.debug("Ignored message: %s", i)
            self.ignore_messages.remove(i)
            return

        contact: LegacyContactType = self.contacts.by_stanza(m)

        url = m["oob"]["url"]
        text = m["body"]
        if (
            m["feature_fallback"]["for"] == self.xmpp["xep_0461"].namespace
            and contact.REPLIES
        ):
            text = m["feature_fallback"].get_stripped_body()

        if url:
            legacy_msg_id = await self.send_file(
                url, contact, reply_to_msg_id=m["reply"]["id"] or None
            )
        elif text:
            legacy_msg_id = await self.send_text(
                text, contact, reply_to_msg_id=m["reply"]["id"] or None
            )
        else:
            log.debug("Ignoring %s", m)
            return
        if legacy_msg_id is not None:
            self.sent[legacy_msg_id] = m.get_id()

    @ignore_message_to_component
    async def active_from_msg(self, m: Message):
        """
        Meant to be called from :class:`BaseGateway` only.

        :param m:
        :return:
        """
        if m.get_to() != self.xmpp.boundjid.bare:
            await self.active(self.contacts.by_stanza(m))

    @ignore_message_to_component
    async def inactive_from_msg(self, m: Message):
        """
        Meant to be called from :class:`BaseGateway` only.

        :param m:
        :return:
        """
        if m.get_to() != self.xmpp.boundjid.bare:
            await self.inactive(self.contacts.by_stanza(m))

    @ignore_message_to_component
    async def composing_from_msg(self, m: Message):
        """
        Meant to be called from :class:`BaseGateway` only.

        :param m:
        :return:
        """
        if m.get_to() != self.xmpp.boundjid.bare:
            await self.composing(self.contacts.by_stanza(m))

    @ignore_message_to_component
    async def paused_from_msg(self, m: Message):
        """
        Meant to be called from :class:`BaseGateway` only.

        :param m:
        :return:
        """
        if m.get_to() != self.xmpp.boundjid.bare:
            await self.paused(self.contacts.by_stanza(m))

    @ignore_message_to_component
    async def displayed_from_msg(self, m: Message):
        """
        Meant to be called from :class:`BaseGateway` only.

        :param m:
        :return:
        """
        displayed_msg_id = m["displayed"]["id"]
        try:
            legacy_msg_id = self.xmpp_msg_id_to_legacy_msg_id(displayed_msg_id)
        except NotImplementedError:
            log.debug("Couldn't convert xmpp msg ID to legacy ID, ignoring read mark")
            return

        await self.displayed(legacy_msg_id, self.contacts.by_stanza(m))

    @ignore_message_to_component
    async def correct_from_msg(self, m: Message):
        xmpp_id = m["replace"]["id"]
        legacy_id = self.sent.inverse.get(xmpp_id)
        if legacy_id is None:
            log.debug("Did not find legacy ID to correct")
            new_legacy_msg_id = await self.send_text(
                m["body"], self.contacts.by_stanza(m)
            )
        else:
            new_legacy_msg_id = await self.correct(
                m["body"], legacy_id, self.contacts.by_stanza(m)
            )
        if new_legacy_msg_id is not None:
            self.sent[new_legacy_msg_id] = m.get_id()

    @ignore_message_to_component
    async def react_from_msg(self, m: Message):
        react_to = m["reactions"]["id"]
        if (legacy_id := self.sent.inverse.get(react_to)) is None:
            log.debug("Cannot find the XMPP ID of this msg: %s", react_to)
            try:
                legacy_id = self.xmpp_msg_id_to_legacy_msg_id(react_to)
            except ValueError:
                log.warning(
                    "Could not convert legacy ID, xmpp reaction was not sent: %s", m
                )
                return
        await self.react(
            legacy_id, [r["value"] for r in m["reactions"]], self.contacts.by_stanza(m)
        )

    @ignore_message_to_component
    async def retract_from_msg(self, m: Message):
        xmpp_id = m["apply_to"]["id"]
        if (legacy_id := self.sent.inverse.get(xmpp_id)) is None:
            log.debug(
                "Cannot find the XMPP ID of this msg: %s, cannot retract", xmpp_id
            )
            return
        await self.retract(legacy_id, self.contacts.by_stanza(m))

    def send_gateway_status(
        self,
        status: Optional[str] = None,
        show=Optional[Literal["away", "chat", "dnd", "xa"]],
        **kwargs,
    ):
        """
        Send a presence from the gateway to the user.

        Can be used to indicate the user session status, ie "SMS code required", "connected", …

        :param status: A status message
        :param show: Presence stanza 'show' element. I suggest using "dnd" to show
            that the gateway is not fully functional
        """
        self.xmpp.send_presence(
            pto=self.user.bare_jid, pstatus=status, pshow=show, **kwargs
        )

    def send_gateway_message(self, text, **msg_kwargs):
        """
        Send a message from the gateway component to the user.

        Can be used to indicate the user session status, ie "SMS code required", "connected", …

        :param text: A text
        """
        self.xmpp.send_message(
            mto=self.user.jid, mbody=text, mfrom=self.xmpp.boundjid, **msg_kwargs
        )

    async def input(self, text: str, **msg_kwargs):
        """
        Request user input via direct messages.

        Wraps call to :meth:`.BaseSession.input`

        :param text: The prompt to send to the user
        :param msg_kwargs: Extra attributes
        :return:
        """
        return await self.xmpp.input(self.user.jid, text, **msg_kwargs)

    async def send_qr(self, text: str):
        """
        Sends a QR code generated from 'text' via HTTP Upload and send the URL to
        ``self.user``

        :param text: Text to encode as a QR code
        """
        await self.xmpp.send_qr(text, mto=self.user.jid)

    def post_init(self):
        """
        Add useful attributes for your session here, if you wish.

        In most cases, this is the right place to add a legacy network-specific
        ``LegacyClient``-like instance attached to this gateway user.
        """
        pass

    async def login(self) -> Optional[str]:
        """
        Login the gateway user to the legacy network.

        Triggered when the gateway start and on user registration.
        It is recommended that this function returns once the user is logged in,
        so if you need to await forever (for instance to listen to incoming events),
        it's a good idea to wrap your listener in an asyncio.Task.

        :return: Optionally, a text to use as the gateway status, e.g., "Connected as 'dude@legacy.network'"
        """
        raise NotImplementedError

    async def logout(self):
        """
        Logout the gateway user from the legacy network.

        Called on user unregistration and gateway shutdown.
        """
        raise NotImplementedError

    async def send_text(
        self,
        t: str,
        c: LegacyContactType,
        *,
        reply_to_msg_id: Optional[LegacyMessageType] = None,
    ) -> Optional[LegacyMessageType]:
        """
        Triggered when the user sends a text message from xmpp to a bridged contact, e.g.
        to ``translated_user_name@slidge.example.com``.

        Override this and implement sending a message to the legacy network in this method.

        :param t: Content of the message
        :param c: Recipient of the message
        :param reply_to_msg_id:
        :return: An ID of some sort that can be used later to ack and mark the message
            as read by the user
        """
        raise NotImplementedError

    async def send_file(
        self,
        u: str,
        c: LegacyContactType,
        *,
        reply_to_msg_id: Optional[LegacyMessageType] = None,
    ) -> Optional[LegacyMessageType]:
        """
        Triggered when the user has sends a file using HTTP Upload (:xep:`0363`)

        :param u: URL of the file
        :param c: Recipient of the file
        :param reply_to_msg_id:
        :return: An ID of some sort that can be used later to ack and mark the message
            as read by the user
        """
        raise NotImplementedError

    async def active(self, c: LegacyContactType):
        """
        Triggered when the user sends an 'active' chat state to the legacy network (:xep:`0085`)

        :param c: Recipient of the active chat state
        """
        raise NotImplementedError

    async def inactive(self, c: LegacyContactType):
        """
        Triggered when the user sends an 'inactive' chat state to the legacy network (:xep:`0085`)

        :param c:
        """
        raise NotImplementedError

    async def composing(self, c: LegacyContactType):
        """
        Triggered when the user starts typing in the window of a legacy contact (:xep:`0085`)

        :param c:
        """
        raise NotImplementedError

    async def paused(self, c: LegacyContactType):
        """
        Triggered when the user pauses typing in the window of a legacy contact (:xep:`0085`)

        :param c:
        """
        raise NotImplementedError

    async def displayed(self, legacy_msg_id: Any, c: LegacyContactType):
        """
        Triggered when the user reads a message sent by a legacy contact.  (:xep:`0333`)

        This is only possible if a valid ``legacy_msg_id`` was passed when transmitting a message
        from a contact to the user in :meth:`.LegacyContact.sent_text` or :meth:`slidge.LegacyContact.send_file`.

        :param legacy_msg_id: Identifier of the message, passed to :meth:`slidge.LegacyContact.send_text`
            or :meth:`slidge.LegacyContact.send_file`
        :param c:
        """
        raise NotImplementedError

    async def correct(
        self, text: str, legacy_msg_id: Any, c: LegacyContactType
    ) -> Optional[LegacyMessageType]:
        """
        Triggered when the user corrected a message using :xep:`0308`

        This is only possible if a valid ``legacy_msg_id`` was passed when transmitting a message
        from a contact to the user in :meth:`.LegacyContact.send_text` or :meth:`slidge.LegacyContact.send_file`.

        :param text:
        :param legacy_msg_id:
        :param c:
        """
        raise NotImplementedError

    async def search(self, form_values: dict[str, str]) -> "SearchResult":
        """
        Triggered when the user uses Jabber Search (:xep:`0055`) on the component

        Form values is a dict in which keys are defined in :attr:`.BaseGateway.SEARCH_FIELDS`

        :param form_values: search query, defined for a specific plugin by overriding
            in :attr:`.BaseGateway.SEARCH_FIELDS`
        :return:
        """
        raise NotImplementedError

    async def react(self, legacy_msg_id: Any, emojis: list[str], c: LegacyContactType):
        """
        Triggered when the user sends message reactions (:xep:`0444`).

        :param legacy_msg_id: ID of the message the user reacts to
        :param emojis: Unicode characters representing reactions to the message ``legacy_msg_id``.
            An empty string means "no reaction", ie, remove all reactions if any were present before
        :param c: Contact the reaction refers to
        """
        raise NotImplementedError

    async def retract(self, legacy_msg_id: Any, c: LegacyContactType):
        """
        Triggered when the user retracts (:xep:`0424`) a message.

        :param legacy_msg_id: Legacy ID of the retracted message
        :param c: The contact this retraction refers to
        """
        raise NotImplementedError


SessionType = TypeVar("SessionType", bound=BaseSession)

_sessions: dict[GatewayUser, BaseSession] = {}
log = logging.getLogger(__name__)



"""
Slidge can be configured via CLI args, environment variables and/or INI files.
To use env vars, use this convention: ``--home-dir`` becomes ``HOME_DIR``.
"""
import importlib
import logging
from pathlib import Path

import configargparse

from slidge import BaseGateway
from slidge.util.db import user_store


# noinspection PyUnresolvedReferences
def get_parser():
    p = configargparse.ArgParser(
        default_config_files=["/etc/slidge/conf.d/*.conf"], description=__doc__
    )
    p.add(
        "-c",
        "--config",
        help="Path to a INI config file.",
        env_var="SLIDGE_CONFIG",
        is_config_file=True,
    )
    p.add(
        "--legacy-module",
        help="Importable python module containing (at least) "
        "a BaseGateway and a LegacySession subclass",
        env_var="SLIDGE_LEGACY_MODULE",
    )
    p.add(
        "-s",
        "--server",
        env_var="SLIDGE_SERVER",
        required=True,
        help="The XMPP server's host name.",
    )
    p.add(
        "-p",
        "--port",
        default="5347",
        env_var="SLIDGE_PORT",
        help="The XMPP server's port for incoming component connections",
    )
    p.add(
        "--secret",
        required=True,
        env_var="SLIDGE_SECRET",
        help="The gateway component's secret (required to connect to the XMPP server)",
    )
    p.add(
        "-j",
        "--jid",
        required=True,
        env_var="SLIDGE_JID",
        help="The gateway component's JID",
    )
    p.add(
        "--upload-service",
        env_var="SLIDGE_UPLOAD",
        help="JID of an HTTP upload service the gateway can use. "
        "This is optional, as it should be automatically determined via service discovery",
    )
    p.add(
        "--home-dir",
        env_var="SLIDGE_HOME_DIR",
        help="Shelve file used to store persistent user data. "
        "Defaults to /var/lib/slidge/${SLIDGE_JID}",
    )
    p.add(
        "--admins",
        env_var="SLIDGE_ADMINS",
        nargs="*",
        help="JIDs of the gateway admins",
    )
    p.add(
        "--user-jid-validator",
        env_var="SLIDGE_RESTRICT",
        help="Regular expression to restrict user that can register to the gateway by JID. "
        "Defaults to .*@${SLIDGE_SERVER}, forbids the gateway to JIDs "
        "not using the same XMPP server as the gateway",
    )
    p.add(
        "--secret-key",
        env_var="SLIDGE_SECRET_KEY",
        help="Encryption for disk storage",
    )
    p.add(
        "--no-roster-push",
        env_var="SLIDGE_NO_ROSTER_PUSH",
        help="Do not fill users' rosters with legacy contacts automatically",
        action="store_true",
    )
    p.add_argument(
        "-q",
        "--quiet",
        help="loglevel=WARNING",
        action="store_const",
        dest="loglevel",
        const=logging.WARNING,
        default=logging.INFO,
        env_var="SLIDGE_QUIET",
    )
    p.add_argument(
        "-d",
        "--debug",
        help="loglevel=DEBUG",
        action="store_const",
        dest="loglevel",
        const=logging.DEBUG,
        env_var="SLIDGE_DEBUG",
    )

    return p


def main():
    args, argv = get_parser().parse_known_args()
    logging.basicConfig(level=args.loglevel)

    if args.home_dir is None:
        args.home_dir = Path("/var/lib/slidge") / args.jid
    if not args.home_dir.exists():
        logging.info("Creating directory '%s'", args.home_dir)
        args.home_dir.mkdir()

    if args.user_jid_validator is None:
        args.user_jid_validator = ".*@" + args.server

    db_file = Path(args.home_dir) / "slidge.db"
    user_store.set_file(db_file, args.secret_key)

    importlib.import_module(args.legacy_module)
    gateway = BaseGateway.get_unique_subclass()(args)
    gateway.config(argv)
    gateway.connect()

    try:
        gateway.loop.run_forever()
    except KeyboardInterrupt:
        logging.debug("Received SIGINT")
    except Exception as e:
        logging.exception(e)
    finally:
        if gateway.is_connected():
            logging.debug("Gateway is connected, cleaning up")
            gateway.shutdown()
            gateway.disconnect()
            gateway.loop.run_until_complete(gateway.disconnected)
        else:
            logging.debug("Gateway is not connected, no need to clean up")
        logging.info("Successful clean shut down")


if __name__ == "__main__":
    main()
